; Simple App Builder

(require 'xmts)

; current widget type to add
(define current-type #f)
(define (set-default-widget widget type) (set! current-type type))

; testing a bitmask
(define (test-bit val bit)
  (let ((final #f))
    (let loop ((val val) (bit bit))
      (if (> bit 1)
	  (loop (quotient val 2) (- bit 1))
	  (set! final val)))
    (odd? final)))

; change property of a widget
(define (change-property widget)
  #t)

(define ready-to-move #f)

(define (really-move widget x y)
  (let ((origx (xt:get-value widget xm:n-x xt:integer))
	(origy (xt:get-value widget xm:n-y xt:integer)))
;    (format #t "moving from (~D,~D) to (~D,~D)\n"
;	    origx origy x y)
    (xt:set-values widget
		   xm:n-x (+ origx x)
		   xm:n-y (+ origy y))))

  
; button pressed on a widget
(define (handle-widget widget event)
  
  (let* ((button (x:get-event-field event x:button-event:button))
	 (type   (x:get-event-field event x:button-event:type))
	 (state  (x:get-event-field event x:button-event:state)))
    (if (= button 3)
	(cond ((test-bit state 3)
;	       (format #t "ControlMask is on.\n")
	       (change-property widget))
	      ((= type x:button-press)
	       (set! ready-to-move #t))
	      ((= type x:button-release)
	       (set! ready-to-move #f))
	      (else
	       (error "Unknown event type."))))))

(define (move-widget widget event)
  (if ready-to-move
      (let ((x (x:get-event-field event x:motion-event:x))
	    (y (x:get-event-field event x:motion-event:y)))
	(really-move widget x y))))
      

; Main window  event handlers
(define (main-key-pressed widget event)
  #t)

(define (main-button-pressed widget event)
  (if current-type
      (let* ((button (x:get-event-field event x:button-event:button))
	     (x      (x:get-event-field event x:button-event:x))
	     (y      (x:get-event-field event x:button-event:y)))
	(if (= button 1)
	    (let ((new (xt:create-managed-widget "silly" (cdr current-type)
						 widget
						 xm:n-x x
						 xm:n-y y)))
	      (xt:add-event-handler new (+ x:button-press-mask
					   x:button-release-mask)
				    0 handle-widget)
	      (xt:add-event-handler new x:pointer-motion-mask 0 
				    move-widget))))))


(xmts:run-app

 "appbuilder"

 (lambda (top)
   (let ((main (xt:create-managed-widget "main" xm:row-column top
					 xm:n-num-columns 2
					 xm:n-packing xm:pack-column))
	 (widget-types `((label         . ,xm:label)
			 (push-button   . ,xm:push-button)
			 (arrow-button  . ,xm:arrow-button)
			 (drawn-button  . ,xm:drawn-button)
			 (list          . ,xm:list)
			 (scale         . ,xm:scale)
			 (scroll-bar    . ,xm:scroll-bar)
			 (text          . ,xm:text)
			 (input-field   . ,xmt:input-field)
			 )))
     (for-each
      (lambda (widget-type)
	(let ((b (xt:create-managed-widget (symbol->string (car widget-type))
					   xm:push-button
					   main)))
	  (xmts:add-activate b set-default-widget widget-type)))
      widget-types)

     (let* ((workshell (xt:create-managed-widget "workshell" 
						xm:dialog-shell main))
	    (work      (xt:create-managed-widget "work"
						 xm:drawing-area workshell)))
       (xt:add-event-handler work x:key-press-mask 0 main-key-pressed)
       (xt:add-event-handler work x:button-press-mask 0 main-button-pressed)
       (xmt:add-delete-callback work xm:unmap 
				(lambda (widget data)
				  (format #t "See ya later! data=~A\n" data)
				  (xt:app-add-time-out work 1000
						       (lambda ()
							 (xt:manage-children
							  work))))
				12345)
))))
				
