;;; Bubble Help

(require 'xmts)
;(require 'format)

;; Give a widget bubble help behavior
(define xmts:add-bubble-help
 
  (let ((bubble-up? #f)
	(last       #f)
	(toid       #f)
	(shell      #f))

    (lambda (widget text timeout)

      ; create bubble help shell
      (define (create-bubble-help-window parent)
	(let* ((shell (xt:create-popup-shell "BubbleHelp" xt:override-shell
					     parent
					     xm:n-allow-shell-resize #t))
	       (label (xt:create-managed-widget "label" xm:label shell
						xm:n-margin-width 10
						xm:n-alignment 
						xm:alignment-center)))
	  (xt:add-event-handler shell x:leave-window-mask 0 
				(lambda (w event) (xt:popdown w)))
	  (cons shell label)))

      ; popup bubble
      (define (show-bubble)
	(if (not shell)
	    (set! shell (create-bubble-help-window widget)))
	(xmts:label-set-string (cdr shell) text)
	(let ((rootxy (xt:translate-coord widget (car last) (cdr last))))
	  (xt:set-values (car shell)
			 xm:n-x (car rootxy)
			 xm:n-y (cdr rootxy)))
	(xt:popup (car shell) 0)
	(x:define-cursor (xt:display (car shell))
			 (xt:window (car shell))
			 xc:crosshair)
	(set! bubble-up? #t))

      ; popdown bubble
      (define (hide-bubble)
	(if bubble-up?
	    (xt:popdown (car shell))))
      
      ; don't count time
      (define (remove-time-out-handler)
	(cond (toid (xt:remove-time-out toid)
		    (set! toid #f))))
      
      ; start counting time
      (define (add-time-out-handler)
	(set! toid (xt:app-add-time-out widget timeout show-bubble)))
      
      ; get current coordinates from event field
      (define (current-coordinates event type)
	(cons (if (eq? type 'crossing)
		  (x:get-event-field event x:crossing-event:x)
		  (x:get-event-field event x:motion-event:x))
	      (if (eq? type 'crossing)
		  (x:get-event-field event x:crossing-event:y)
		  (x:get-event-field event x:motion-event:y))))
      
      ; save current position
      (define (save-position event type)
	(set! last (current-coordinates event type)))
      
      ; moving event handler
      (define (moving widget event)
;	(hide-bubble)
	(save-position event 'motion)
	(remove-time-out-handler)
	(add-time-out-handler))
      
      ; enter window event handler
      (define (entering widget event)
	(save-position event 'crossing)
	(add-time-out-handler))
      
      ; leave window event handler
      (define (leaving widget event) (remove-time-out-handler))

      ; pushing window event handler
      (define (pushing widget event) (remove-time-out-handler))

      ; main body of add-bubble-help
      (xt:add-event-handler widget x:enter-window-mask 0 entering)
      (xt:add-event-handler widget x:leave-window-mask 0 leaving)
      (xt:add-event-handler widget x:pointer-motion-mask 0 moving)
      (xt:add-event-handler widget x:button-press-mask 0 moving))))
  

(define (xmts:add-bubble-help-simple widget text)
  (xmts:add-bubble-help widget text 2000))
