/*
 * $Header: /home/tomkwong/src/xmtscm/RCS/cc.c,v 1.1 1995/12/05 23:06:26 tomkwong Exp $
 *
 * C-Client Library Extension to SCM
 * Copyright 1995 Thomas Kwong
 */

#include "mail.h"
#include "smtp.h"
#include "misc.h"
#include "osdep.h"
#include <Xm/Xm.h>
#include "scm.h"
#include "x.h"
#include "xt.h"

/* From unix.c */
extern char *myhost(), *myname(), *myrealname();

/* flags */
#define NUM_SETTABLE_FLAGS 3
static char *settable_flags[] = {
    "\\DELETED", "\\ANSWERED", "\\FLAGGED",
};

#define MAILSTREAMP(x) (TYP16(x) == tc16_mailstream)
#define GETMAILSTREAM(x) ((MAILSTREAM*) CDR(x))
#define MKBOOL(x) ((x)?BOOL_T:BOOL_F)
#define MKSTR(s) (makfromstr((s),strlen(s)))

/* forward declaration */
long tc16_mailstream;
static sizet free_mailstream();
static int   print_mailstream();
static SCM   equalp_mailstream();

static char s_open_mailstream[]	    = "cc:open-mail-stream";
/*static char 	s_close_mailstream[]	= "cc:close-mail-stream";*/
static char s_expunge_mailstream[]  = "cc:expunge-mail-stream";
static char s_fetch_header[]	    = "cc:fetch-header";
static char s_fetch_text[]	    = "cc:fetch-text";
static char s_fetch_stream_info[]   = "cc:fetch-stream-info";
static char s_fetch_message_info[]  = "cc:fetch-message-info";
static char s_fetch_subject[]	    = "cc:fetch-subject";
static char s_fetch_from[]	    = "cc:fetch-from";
static char s_set_flag[]            = "cc:set-flag";
static char s_clear_flag[]          = "cc:clear-flag";
static char s_copy_message[]        = "cc:copy-message";
static char s_register_hook[]       = "cc:register-hook";
static char s_unregister_hook[]     = "cc:unregister-hook";
static char s_smtp_send_message[]   = "cc:smtp-send-message";
static char s_check[]               = "cc:check";
/* my own procedures */
static char s_reply_address[]       = "cc:reply-address";
static char s_getpass[]             = "cc:getpass";
static char s_set_value[]           = "cc:set!";

#define LOGIN_HOOK         0
#define LOG_HOOK           1
#define NOTIFY_HOOK        2
#define NEWMSG_HOOK        3
#define MAX_CC_HOOKS       4
static SCM cc_hooks[] = {
    0,					/* mm_login */
    0,					/* mm_log   */
    0,					/* mm_notify */
    0,					/* mm_exists with new messages */
};

/* these are settable by cc:set-value */
#define SET_DOMAIN         0
#define SET_NAME           1
#define SETVALUE_MAXLEN    256
#define SETTABLE(i) (INUM(i) >= 0 && INUM(i) < 2)
char values[2][SETVALUE_MAXLEN] = {
    "",
    "",
};

#define SCMLIST_TO_SEQUENCE "scmlist_to_sequence"
#define CALCPAD -9999
#define OUTRANGE "Out of range"

/* Scheme types defined in this module */
static smobfuns mailstream_smob = {
  mark_no_further, 
  free_mailstream,			/* why free_mailstream not working? */
  print_mailstream,
  equalp_mailstream
};

SCM make_mailstream()
{
  SCM s;
  NEWCELL(s);
  CAR(s) = tc16_mailstream;
  CDR(s) = 0;
  return s;
}

sizet free_mailstream(ptr)
SCM ptr;
{
/*  printf("Closing mail stream [%s]\n", GETMAILSTREAM(ptr)->mailbox);*/
  mail_close(GETMAILSTREAM(ptr));
  return sizeof(MAILSTREAM);
}

int print_mailstream(exp, f, writing)
SCM exp;
SCM f;
int writing;
{
  lputs("#<MailStream>", f);
  return 1;
}

SCM equalp_mailstream(s1, s2)
SCM s1, s2;
{
  return (GETMAILSTREAM(s1) == GETMAILSTREAM(s2)) ? BOOL_T : BOOL_F;
}


/************ Export Routines *************/

SCM open_mailstream(path)
SCM path;
{
  MAILSTREAM *stream = NIL;
  SCM s;

  ASSERT(NIMP(path) && STRINGP(path), path, ARG1, s_open_mailstream);
/*  printf("opening mail stream = [%s]\n", CHARS(path));*/
  if ((stream = mail_open(stream, CHARS(path), OP_DEBUG))) {
      s = make_mailstream();
      SETCDR(s, stream);
      return s;
  }
  return BOOL_F;
}

/* Let SCM GC do this
SCM close_mailstream(stream)
SCM stream;
{
  ASSERT(NIMP(stream) && MAILSTREAMP(stream), stream, ARG1, 
	 s_close_mailstream);
  mail_close(GETMAILSTREAM(stream));
  return UNSPECIFIED;
}
*/

SCM expunge_mailstream(stream)
SCM stream;
{
  ASSERT(NIMP(stream) && MAILSTREAMP(stream), stream, ARG1, 
	 s_expunge_mailstream);
  mail_expunge(GETMAILSTREAM(stream));
  return UNSPECIFIED;
}

SCM fetch_header(stream, msgno)
SCM stream, msgno;
{
  char *p;

  ASSERT(NIMP(stream) && MAILSTREAMP(stream), stream, ARG1, s_fetch_header);
  ASSERT(INUMP(msgno) && INUM(msgno) > 0 && 
	 INUM(msgno) <= GETMAILSTREAM(stream)->nmsgs, 
	 msgno, ARG2, s_fetch_header);
  p = mail_fetchheader(GETMAILSTREAM(stream), INUM(msgno));
  return MKSTR(p);
}

SCM fetch_text(stream, msgno)
SCM stream, msgno;
{
  char *p;

  ASSERT(NIMP(stream) && MAILSTREAMP(stream), stream, ARG1, s_fetch_text);
  ASSERT(INUMP(msgno) && INUM(msgno) > 0 && 
	 INUM(msgno) <= GETMAILSTREAM(stream)->nmsgs, 
	 msgno, ARG2, s_fetch_text);
  p = mail_fetchtext(GETMAILSTREAM(stream), INUM(msgno));
  return MKSTR(p);
}

SCM fetch_stream_info(stream)
SCM stream;
{
  MAILSTREAM *s;
  SCM result;

  ASSERT(NIMP(stream) && MAILSTREAMP(stream), stream, ARG1, 
	 s_fetch_stream_info);
  s = GETMAILSTREAM(stream);
  result = cons(MKSTR(s->mailbox), EOL);
  result = cons(MAKINUM(s->recent), result);
  result = cons(MAKINUM(s->nmsgs), result);
  return result;
}

SCM fetch_message_info(stream, msgno)
SCM stream, msgno;
{
  MAILSTREAM *s;
  ENVELOPE *env;
  BODY *body;
  MESSAGECACHE *cache;
  SCM result;

  ASSERT(NIMP(stream) && MAILSTREAMP(stream), stream, ARG1, 
	 s_fetch_message_info);
  s = GETMAILSTREAM(stream);
  ASSERT(INUMP(msgno) && INUM(msgno) > 0 && INUM(msgno) <= s->nmsgs, 
	 msgno, ARG2, s_fetch_message_info);

  env = mail_fetchstructure(s, INUM(msgno), &body);
  cache = mail_elt(s, INUM(msgno));
  result = cons(MAKINUM(body->size.lines), EOL);
  result = cons(MAKINUM(body->size.bytes), result);
  result = cons(MKBOOL(cache->recent), result);
  result = cons(MKBOOL(cache->seen), result);
  result = cons(MKBOOL(cache->flagged), result);
  result = cons(MKBOOL(cache->answered), result);
  result = cons(MKBOOL(cache->deleted), result);

  return result;
}

SCM fetch_subject(stream, msgno, args)
SCM stream, msgno, args;
{
  MAILSTREAM *s;
  SCM len;
  char subject[1024];

  ASSERT(NIMP(stream) && MAILSTREAMP(stream), stream, ARG1, s_fetch_subject);
  s = GETMAILSTREAM(stream);
  ASSERT(INUMP(msgno) && INUM(msgno) > 0 && INUM(msgno) <= s->nmsgs, 
	 msgno, ARG2, s_fetch_subject);

  ASSERT(NIMP(args) && CONSP(args), len, ARG3, s_fetch_subject);
  len = CAR(args); 
  ASSERT(INUMP(len) && INUM(len) > 0 && INUM(len) < 1024, len, ARG3, 
	 s_fetch_subject);

  mail_fetchsubject(subject, s, INUM(msgno), INUM(len));
  return MKSTR(subject);
}

SCM fetch_from(stream, msgno, args)
SCM stream, msgno, args;
{
  MAILSTREAM *s;
  SCM len;
  char from[1024];

  ASSERT(NIMP(stream) && MAILSTREAMP(stream), stream, ARG1, s_fetch_from);
  s = GETMAILSTREAM(stream);
  ASSERT(INUMP(msgno) && INUM(msgno) > 0 && INUM(msgno) <= s->nmsgs, 
	 msgno, ARG2, s_fetch_from);

  ASSERT(!NULLP(args), len, ARG3, s_fetch_from);
  len = CAR(args); 
  ASSERT(INUMP(len) && INUM(len) > 0 && INUM(len) < 1024, len, ARG3, 
	 s_fetch_subject);

  mail_fetchfrom(from, s, INUM(msgno), INUM(len));
  return MKSTR(from);
}

static char *scmlist_to_sequence(list, lower, upper)
SCM list;
int lower, upper;
{
  static char result[1024];
  int n;
  char num[12];

  strcpy(result, "");
  while (list != EOL)
  {
      ASSERT(INUMP(CAR(list)), CAR(list), "*expecting an integer*", 
	     SCMLIST_TO_SEQUENCE);
      n = INUM(CAR(list));
      if (lower != CALCPAD) ASSERT(n >= lower, CAR(list), OUTRANGE,
				   SCMLIST_TO_SEQUENCE);
      if (upper != CALCPAD) ASSERT(n <= upper, CAR(list), OUTRANGE,
				   SCMLIST_TO_SEQUENCE);
      sprintf(num, "%d,", n);
      strcat(result, num);
      list = CDR(list);
  }
  result[strlen(result)-1] = '\0';
/*  fprintf(stderr, "scmlist_to_sequence: result=[%s]\n", result);*/
  return result;
}

static SCM set_or_unset_flag(stream, msgs, args, set)
SCM stream, msgs, args, set;
{
  MAILSTREAM *s;
  SCM flag;
  char *seq;
  char tmp[10];

  ASSERT(NIMP(stream) && MAILSTREAMP(stream), stream, ARG1, s_set_flag);
  s = GETMAILSTREAM(stream);

  if (INUMP(msgs)) {			/* just one message */
      int num = INUM(msgs);
      if (num > 0 && num <= s->nmsgs) {
	  seq = tmp;
	  sprintf(seq, "%d", num);
      } else 
	  return BOOL_F;
  } else {				/* list of messages */
      seq = scmlist_to_sequence(msgs, 1, s->nmsgs);
  }

  ASSERT(NIMP(args) && CONSP(args), args, ARG3, s_set_flag);
  flag = CAR(args);
  ASSERT(INUMP(flag) && INUM(flag) >= 0 && 
	 INUM(flag) < NUM_SETTABLE_FLAGS, 
	 flag, ARG3, s_set_flag);
  
  if (set == BOOL_T)
      mail_setflag(s, seq, settable_flags[INUM(flag)]);
  else
      mail_clearflag(s, seq, settable_flags[INUM(flag)]);

  return BOOL_T;
}

SCM set_flag(stream, msgs, args)
SCM stream, msgs, args;
{
  return set_or_unset_flag(stream, msgs, args, BOOL_T);
}

SCM clear_flag(stream, msgs, args)
SCM stream, msgs, args;
{
  return set_or_unset_flag(stream, msgs, args, BOOL_F);
}

SCM copy_message(stream, msglist, args)
SCM stream, msglist, args;
{
  MAILSTREAM *s;
  SCM dest;

  ASSERT(NIMP(stream) && MAILSTREAMP(stream), stream, ARG1, s_copy_message);
  ASSERT(NIMP(msglist) && CONSP(msglist), msglist, ARG2, s_copy_message);
  ASSERT(NIMP(args) && CONSP(args), args, ARG3, s_copy_message);
  dest = CAR(args);
  ASSERT(NIMP(dest) && STRINGP(dest), dest, ARG3, s_copy_message);

  s = GETMAILSTREAM(stream);
  if (mail_copy(s, scmlist_to_sequence(msglist, 1, s->nmsgs),
		CHARS(dest)))
      return BOOL_T;
  else
      return BOOL_F;
}

SCM register_hook(type, hook)
SCM type, hook;
{
  ASSERT(INUMP(type) && INUM(type) >= 0 && INUM(type) < MAX_CC_HOOKS,
	 type, ARG1, s_register_hook);
  ASSERT(NIMP(hook) && CLOSUREP(hook), hook, ARG2, s_register_hook);
  cc_hooks[INUM(type)] = hook;
  return UNSPECIFIED;
}

SCM unregister_hook(type)
SCM type;
{
  ASSERT(INUMP(type) && INUM(type) >= 0 && INUM(type) < MAX_CC_HOOKS,
	 type, ARG1, s_unregister_hook);
  cc_hooks[INUM(type)] = 0;
  return UNSPECIFIED;
}

/* Internal */
char **construct_hosts(list, size)
SCM list;
int *size;
{
  int len, i;
  SCM ptr;
  char **ar;

  for (len = 0, ptr = list; ptr != EOL; len++, ptr = CDR(ptr))
      ;
/*  printf("construct_hosts: length of list=%d\n", len);*/

  ar = (char **) malloc(sizeof(char*) * (len + 1));
  for (len = 0, ptr = list; ptr != EOL; len++, ptr = CDR(ptr)) {
      ASSERT(NIMP(CAR(ptr)) && STRINGP(CAR(ptr)), CAR(ptr),
	     "*expecting a string*", "construct_hosts");
      ar[len] = strdup(CHARS(CAR(ptr)));
  }
  ar[len] = NIL;

/*  for (i = 0; i < len; i++)*/
/*      printf("ar[%d] = '%s'\n", i, ar[i]);*/

  *size = len;
  return ar;
}

/* (cc:smtp-send-message host to subject cc body) */
/* The first 3 arguments are extra info */
SCM smtp_send_message(hostlist, to, args)
{
  SCM subject, cc, sbody;
  char line[MAILTMPLEN];
  SMTPSTREAM *stream = NIL;
  ENVELOPE *env = mail_newenvelope();
  BODY *body = mail_newbody();
  SCM result = BOOL_F;
  char *personal, *domain, *login, **hosts;
  int i, hlen;

  ASSERT(NIMP(hostlist) && CONSP(hostlist), hostlist, ARG1, 
	 s_smtp_send_message);
  hosts = construct_hosts(hostlist, &hlen);

  ASSERT(NIMP(to) && STRINGP(to), to, ARG2, s_smtp_send_message);

  ASSERT(NIMP(args) && CONSP(args), args, ARG3, s_smtp_send_message);
  subject = CAR(args); args = CDR(args);
  ASSERT(NIMP(subject) && STRINGP(subject), subject, ARG3, 
	 s_smtp_send_message);

  ASSERT(NIMP(args) && CONSP(args), args, ARG4, s_smtp_send_message);
  cc = CAR(args);args = CDR(args);
  ASSERT(NIMP(cc) && STRINGP(cc), cc, ARG4, s_smtp_send_message);

  ASSERT(NIMP(args) && CONSP(args), args, ARG5, s_smtp_send_message);
  sbody = CAR(args);args = CDR(args);
  ASSERT(NIMP(sbody) && STRINGP(sbody), sbody, ARG5, 
	 s_smtp_send_message);

/*  hostlist[0] = cpystr(CHARS(host));*/
/*  hostlist[1] = NIL;*/

  personal = strlen(values[SET_NAME]) ? values[SET_NAME] : myrealname();
  domain   = strlen(values[SET_DOMAIN]) ? values[SET_DOMAIN] : myhost();
  login    = myname();
/*  printf("Send: personal=[%s], domain=[%s]\n", personal, domain);*/
  
  env->from = mail_newaddr();
  env->from->personal = cpystr(personal);
  env->from->mailbox  = cpystr(login);
  env->from->host     = cpystr(domain);

  env->return_path = mail_newaddr();
  env->return_path->personal = cpystr(personal);
  env->return_path->mailbox  = cpystr(login);
  env->return_path->host     = cpystr(domain);

  rfc822_parse_adrlist(&env->to, CHARS(to), domain);
  rfc822_parse_adrlist(&env->cc, CHARS(cc), domain);
  env->subject = cpystr(CHARS(subject));

  body->type = TYPETEXT;
  body->contents.text = cpystr(CHARS(sbody));
  rfc822_date(line);
  env->date = (char*) fs_get(1+strlen(line));
  strcpy(env->date, line);

/*  printf("personal=[%s], mailbox=[%s], host=[%s]\n", */
/*	 env->return_path->personal,*/
/*	 env->return_path->mailbox,*/
/*	 env->return_path->host);*/
/*  printf("To: %s\n", CHARS(to));*/
/*  printf("Subject: %s\n", CHARS(subject));*/
/*  printf("Cc: %s\n", CHARS(cc));*/
/*  printf("\b%s\n", CHARS(sbody));*/

  if (env->to) {
      if (stream = smtp_open(hosts, T)) 
      {
	  if (smtp_mail(stream, "MAIL", env, body))
	      result = BOOL_T;
	  else
	      result = BOOL_F;
	  smtp_close(stream);
      }
      mail_free_envelope(&env);
      mail_free_body(&body);
  }
/*  free(hostlist[0]);*/
  for (i = 0; i < hlen; i++)
      free(hosts[i]);
  free(hosts);

  return result;
}

SCM check(stream)
SCM stream;
{
  ASSERT(NIMP(stream) && MAILSTREAMP(stream), stream, ARG1, s_reply_address);
  mail_check(GETMAILSTREAM(stream));
  return UNSPECIFIED;
}

SCM reply_address(stream, msgno)
SCM stream, msgno;
{
  MAILSTREAM *s;
  ENVELOPE *env;
  ADDRESS *p;
  char addr[1024];

  ASSERT(NIMP(stream) && MAILSTREAMP(stream), stream, ARG1, s_reply_address);
  s = GETMAILSTREAM(stream);
  ASSERT(INUMP(msgno) && INUM(msgno) > 0 && INUM(msgno) <= s->nmsgs , 
	 msgno, ARG2, s_reply_address);
  env = mail_fetchstructure(s, INUM(msgno), NIL);
  p = env->reply_to ? env->reply_to : env->from;
  if (p && p->personal)
      sprintf(addr, "%s <%s@%s>", p->personal, p->mailbox, p->host);
  else if (p)
      sprintf(addr, "%s@%s", p->mailbox, p->host);
  else
      sprintf(addr, "");
  return MKSTR(addr);
}

/* Useful for text-based clients */
SCM mygetpass(prompt)
SCM prompt;
{
  char *pass;

  ASSERT(NIMP(prompt) && STRINGP(prompt), prompt, ARG1, s_getpass);
  pass = getpass(CHARS(prompt));
  return MKSTR(pass);
}

SCM set_value(which, str)
SCM which, str;
{
  ASSERT(INUMP(which) && SETTABLE(which), which, ARG1, s_set_value);
  ASSERT(NIMP(str) && STRINGP(str) && strlen(CHARS(str)) < SETVALUE_MAXLEN,
	 str, ARG2, s_set_value);

  strcpy(values[INUM(which)], CHARS(str));
/*  printf("values[%d]=%s\n", INUM(which), values[INUM(which)]);*/
  return UNSPECIFIED;
}

/************** Routine Specifications ************/

static iproc subr1s[] = {
  {s_open_mailstream,	open_mailstream},
/*  {s_close_mailstream,  close_mailstream},*/
  {s_fetch_stream_info,	fetch_stream_info},
  {s_expunge_mailstream,expunge_mailstream},
  {s_getpass,           mygetpass},
  {s_check,             check},
  {s_unregister_hook,   unregister_hook},
  {0,0}
};


static iproc subr2s[] = {
  {s_fetch_header,	fetch_header},
  {s_fetch_text,	fetch_text},
  {s_fetch_message_info,fetch_message_info},
  {s_register_hook,     register_hook},
  {s_reply_address,     reply_address},
  {s_set_value,         set_value},
  {0,0}
};

static iproc lsubr2s[] = {
  {s_fetch_subject,	fetch_subject},
  {s_fetch_from,	fetch_from},
  {s_set_flag,		set_flag},
  {s_clear_flag,	clear_flag},
  {s_copy_message,      copy_message},
  {s_smtp_send_message, smtp_send_message},
  {0,0}
};

void init_cc()
{
  init_iprocs(subr1s,  tc7_subr_1);
  init_iprocs(subr2s,  tc7_subr_2);
  init_iprocs(lsubr2s, tc7_lsubr_2);
  tc16_mailstream = newsmob(&mailstream_smob);
  mail_link (&imapdriver);            /* link in the imap driver */
  mail_link (&nntpdriver);            /* link in the nntp driver */
  mail_link (&pop3driver);            /* link in the pop3 driver */
  mail_link (&mhdriver);              /* link in the mh driver */
  mail_link (&mtxdriver);             /* link in the mtx driver */
  mail_link (&tenexdriver);           /* link in the tenex driver */
  mail_link (&mmdfdriver);            /* link in the mmdf driver */
  mail_link (&bezerkdriver);          /* link in the bezerk driver */
  mail_link (&newsdriver);            /* link in the news driver */
  mail_link (&philedriver);           /* link in the phile driver */
  mail_link (&dummydriver);           /* link in the dummy driver */
}


/* here are the c-client library stubs */
void
mm_searched(MAILSTREAM *stream, long number)
{
/*    printf("searched: %ld", number);*/
}

void
mm_exists(MAILSTREAM *stream, long number)
{
/*    printf("exists: lock=%d, number=%ld\n", stream->lock, number);*/
/*    if (!stream->lock && number > stream->nmsgs) /* new message arrived */
/*    printf("number=%d, stream->nmsgs=%d\n", number, stream->nmsgs);*/
    if (number > stream->nmsgs)
    {
	if (cc_hooks[NEWMSG_HOOK])
	    apply(cc_hooks[NEWMSG_HOOK], MAKINUM(number), listofnull);
/*	else*/
/*	    printf("exists: %ld\n", number);*/
    }
/*    printf("exists: %ld", number);*/
}

void
mm_expunged(MAILSTREAM *stream, long number)
{
/*    printf("expunge: %ld", number);*/
}
void
mm_mailbox(char *string)
{
/*    printf("mailbox: %s\n", string);*/
}

void
mm_bboard(char *string)
{
/*    printf("bboard: %s\n", string);*/
}

void
mm_notify(MAILSTREAM *stream, char *string, long errflag)
{
/*    printf("notify: %s (errflag=%d, locked=%d)\n", string, errflag, */
/*	   stream->lock);*/
    if (!stream->lock && cc_hooks[NOTIFY_HOOK])
	apply(cc_hooks[NOTIFY_HOOK], cons(makfromstr(string,strlen(string)),
					  MAKINUM(errflag)),
	      listofnull);
}

void
mm_log(char *string, long errflag)
{
/*    printf("log: %s (%d)\n", string, errflag);*/
    if (cc_hooks[LOG_HOOK])
	apply(cc_hooks[LOG_HOOK], cons(makfromstr(string,strlen(string)),
				       MAKINUM(errflag)),
	      listofnull);
}

void
mm_dlog(char *string)
{
/*    mm_log(string, (long)9999);*/
}

void
mm_login(char *host, char *user, char *pwd, long trial)
{
    SCM result, suser, spwd;
    
    if (cc_hooks[LOGIN_HOOK])
    {
/*	printf("host=[%s]\n",host);*/
	result = apply(cc_hooks[LOGIN_HOOK], makfromstr(host, strlen(host)), 
		       listofnull);
	ASSERT(NIMP(result) && CONSP(result), result, 
	       "Wrong result (must be a pair of strings)",
	       "mm_login");
	suser = CAR(result); 
	spwd  = CDR(result);
	ASSERT(NIMP(suser) && STRINGP(suser), suser, 
	       "Type mismatch for username", "mm_login");
	ASSERT(NIMP(spwd) && STRINGP(spwd), spwd, 
	       "Type mismatch for password", "mm_login");
	strcpy(user, CHARS(suser));
	strcpy(pwd,  CHARS(spwd));
    }
    else
    {
	printf("Enter username: "); fflush(stdout);
	scanf("%s", user);
	strcpy(pwd, getpass("Enter password: "));
    }
}

void
mm_critical(MAILSTREAM *stream)
{
/*    printf("mm_critical\n");*/
}

void
mm_nocritical(MAILSTREAM *stream)
{
/*    printf("mm_nocritical\n");*/
}

long
mm_diskerror(MAILSTREAM *stream, long errcode, long serious)
{
    fprintf(stderr, "diskerror %ld\n", errcode);
    return (long)T;
}
void
mm_fatal(char *string)
{
/*    mm_log(string, ERROR);*/
}

void
mm_flags(MAILSTREAM *stream, long number)
{
/*    printf("mm_flags, number=%ld\n", number);*/
}
