;;; Xmscm version of controls.c in CDE example.

(require 'xm)
(require 'xmconst)
(require 'dt)
(require 'dtconst)
(require 'xmts)

(define (main)

  ; Convenience routine(?)
  (define (c proc parent name . args)
    (let ((widget (apply proc (append (list parent name) args))))
      (xt:manage-children widget)
      widget))

  (define (create-spin-boxes parent)
    (let* ((values #("alpha" "beta" "gamma" "delta"
		     "epsilon" "zeta" "eta" "theta"
		     "iota" "kappa" "lambda" "mu"
		     "nu" "xi" "omicron" "pi"
		     "rho" "sigma" "tau" "upsilon"
		     "phi" "chi" "psi" "omega"))
	   (xmvec   (xmts:vector-map xm:string-create values))
	   (xmtab   (xm:vector->xmstringtable xmvec))
	   (len     (vector-length values)))

      ; Create title label
      (xt:create-managed-widget "title" xm:label parent
				xm:n-label-string
				(xm:string-create "SpinBox Widget"))
      
      (xt:create-managed-widget "spinBox1" dt:spin-box parent
				dt:n-values xmtab
				dt:n-num-values len
				dt:n-columns 10)

      (xt:create-managed-widget "spinBox2" dt:spin-box parent
				dt:n-spin-box-child-type dt:numeric
				dt:n-minimum-value 1000
				dt:n-maximum-value 100000
				dt:n-increment-value 1000
				dt:n-decimal-points 3
				dt:n-position 1000
				dt:n-arrow-layout dt:arrows-split
				dt:n-columns 10)

      (xt:add-callback
       (xt:create-managed-widget "spinBox3" dt:spin-box parent
				 dt:n-spin-box-child-type dt:numeric
				 dt:n-minimum-value 1500
				 dt:n-maximum-value 60500
				 dt:n-increment-value 1500
				 dt:n-decimal-points 2
				 dt:n-position 7500
				 dt:n-arrow-layout dt:arrows-flat-beginning
				 dt:n-columns 10)
       dt:n-modify-verify-callback
       (let ((allow-change #f))
	 (lambda (widget client-data callback-data)
	   (set! allow-change (if allow-change #f #t))
	   ;(write callback-data) (newline)
	   (if (not allow-change)
	       (begin (format #t
		  "DtSpinBox: DtNmodifyVerifyCallback. Change disallowed.\n")
		      (x:bell (xt:display widget) 100)))
	   allow-change))
       #f)

      (xt:create-managed-widget "spinBox4" dt:spin-box parent
				dt:n-values xmtab
				dt:n-num-values len
				dt:n-arrow-layout dt:arrows-beginning
				dt:n-columns 10)
  
      (xt:create-managed-widget "spinBox5" dt:spin-box parent
				dt:n-spin-box-child-type dt:numeric
				dt:n-minimum-value 1000
				dt:n-maximum-value 100000
				dt:n-increment-value 1000
				dt:n-decimal-points 3
				dt:n-position 1000
				dt:n-arrow-layout dt:arrows-flat-end
				dt:n-columns 10)))

  (define (create-combo-boxes parent)
    (let* ((values #("alpha" "beta" "gamma" "delta"
		     "epsilon" "zeta" "eta" "theta"
		     "iota" "kappa" "lambda" "mu"
		     "nu" "xi" "omicron" "pi"
		     "rho" "sigma" "tau" "upsilon"
		     "phi" "chi" "psi" "omega"))
	   (valstr (xmts:vector-map xm:string-create values))
	   (valtab (xm:vector->xmstringtable valstr))
	   (vallen (vector-length values))
	   (colors #("Red" "Yellow" "Green" "Brown" "Blue"))
	   (colstr (xmts:vector-map xm:string-create colors))
	   (coltab (xm:vector->xmstringtable colstr))
	   (collen (vector-length colors)))

      ; Create title label
      (xt:create-managed-widget "title" xm:label parent
				xm:n-label-string
				(xm:string-create "ComboBox Widget"))

      (let* ((c (xt:create-managed-widget "comboBox1" dt:combo-box parent
					  dt:n-combo-box-type
					  dt:drop-down-combo-box
					  dt:n-items coltab
					  dt:n-item-count collen
					  dt:n-visible-item-count 5
					  dt:n-columns 10))
	     (list   (xt:name->widget c "*List")))
	(xm:list-add-items list valtab 10 0)
	(xt:set-values list xm:n-visible-item-count 10)
	(xt:add-callback c dt:n-selection-callback
			 (lambda (widget client-data callback-data)
			   (write callback-data)
			   (newline))
			 #f))

      (let* ((c (xt:create-managed-widget "comboBox2" dt:combo-box parent
					  dt:n-combo-box-type
					  dt:drop-down-combo-box
					  dt:n-orientation dt:left
					  dt:n-columns 10))
	     (list   (xt:name->widget c "*List")))
	(xm:list-add-items list valtab 7 0)
	(xt:set-values c dt:n-selected-position 3))

      (let ((c (xt:create-managed-widget "comboBox3" dt:combo-box parent
					 dt:n-orientation dt:left
					 dt:n-items valtab
					 dt:n-item-count vallen
					 dt:n-visible-item-count 8)))
	(xt:set-values c dt:n-selected-item (xm:string-create "gamma")))
      
      (let ((c (xt:create-managed-widget "comboBox4" dt:combo-box parent
					 dt:n-orientation dt:right
					 dt:n-visible-item-count 8)))
	(do ((i 0 (+ i 1)))
	    ((= i vallen))
	  (dt:combo-box-add-item c (vector-ref valstr i) 0 #t))
	(xt:set-values c dt:n-selected-position 4))))

  (define (create-menu-buttons parent)

    ; Create title label
    (xt:create-managed-widget "title" xm:label parent
			      xm:n-label-string
			      (xm:string-create "MenuButton Widget"))

    (let ((m (xt:create-managed-widget "menuButton1" dt:menu-button parent
				       xm:n-label-string
				       (xm:string-create "Action"))))
      (let ((submenu (xt:get-value m dt:n-sub-menu-id xt:widget)))
	(xt:create-managed-widget "Push" xm:push-button submenu)
	(xt:create-managed-widget "Pull" xm:push-button submenu)
	(xt:create-managed-widget "Turn" xm:push-button submenu)))

    (let ((m (xt:create-managed-widget "menuButton2" dt:menu-button parent
				       xm:n-label-string
				       (xm:string-create "Movement"))))
      (let ((submenu (xm:create-popup-menu m "submenu"
					   xm:n-tear-off-model 
					   xm:tear-off-enabled)))
	(xt:create-managed-widget "Run" xm:push-button submenu)
	(xt:create-managed-widget "Jump" xm:push-button submenu)
	(xt:create-managed-widget "Stop" xm:push-button submenu)
	(xt:set-values m dt:n-sub-menu-id submenu)
	(let* ((depth (xt:get-value m xm:n-depth xt:integer))
	       (fg    (xt:get-value m xm:n-foreground xt:integer))
	       (bg    (xt:get-value m xm:n-background xt:integer))
	       (pix   (xm:get-pixmap-by-depth m "menu_glyph.xbm" fg bg depth)))
	  (if pix
	      (xt:set-values m dt:n-cascade-pixmap pix)
	      (format #t "Warning: Cannot load pixmap.\n"))))))

  (let* ((toplevel        (xt:app-initialize "controls" "Control"))
	 (main-window     (c xm:create-work-area toplevel "mainWindow"
			     xm:n-orientation xm:horizontal
			     xm:n-spacing 40))
	 (spin-container  (c xm:create-work-area main-window "spinContainer"
			     xm:n-spacing 20))
	 (combo-container (c xm:create-work-area main-window "comboContainer"
			     xm:n-spacing 20))
	 (menu-container  (c xm:create-work-area main-window "menuContainer"
			     xm:n-spacing 20)))
    (create-spin-boxes   spin-container)
    (create-combo-boxes  combo-container)
    (create-menu-buttons menu-container)
    (xt:realize-widget toplevel)
    (xt:app-main-loop toplevel))
  )

(main)
