;;;
;;; Clock -- Calls a procedure at regular time interval.
;;;       -- The procedure must accept a single argument of how many times
;;;          the procedure has been called.
;;;
;;; Copyright 1995 by Thomas Kwong

(define (make-timer widget proc timeout)

  (define handler
    (let ((count 0))
      (lambda ()
	(set! count (+ count 1))
	(if (proc count)
	    (xt:app-add-time-out widget timeout handler)))))
	    
  (xt:app-add-time-out widget timeout handler))

; Testing
(require 'xmts)
(require 'posix-time)
;(set! *xmt-debug-mode* #t)

(xmts:run-app 
 "clock" 
 (lambda (top)

   ; Just some good looking resources
   (xmts:put-line-resources 
    top
    "*fontList: *helvetica-medium-r-normal-*-*-240-*"
    "*background: papayawhip"
    "*foreground: firebrick")

   (let* ((txt (xt:create-managed-widget "txt" xm:text top
					 xm:n-shadow-thickness 0
					 xm:n-cursor-position-visible #f))
	  (msk (xmt:get-bitmap top "icons/hourglass.mask.xbm"))
	  (csr (vector (xmt:get-bitmap top "icons/hourglass_1.xbm")
		       (xmt:get-bitmap top "icons/hourglass_2.xbm")
		       (xmt:get-bitmap top "icons/hourglass_3.xbm")
		       (xmt:get-bitmap top "icons/hourglass_4.xbm")
		       (xmt:get-bitmap top "icons/hourglass_5.xbm")
		       (xmt:get-bitmap top "icons/hourglass_6.xbm")))
	  
	  (dpy (xt:display top)))

     (debug csr msk)

     ; Update clock display every second
     (make-timer txt 
		 (lambda (count)
		   (xm:text-set-string txt 
				       (asctime (localtime (current-time)))))
		 1000)

     ; This one demonstrate a constantly changed cursor
     (make-timer top
		 (lambda (count)
		   (if (xt:is-realized top)
		       (let ((rmd (remainder count 6)))
			 (x:define-pixmap-cursor 
			  dpy (xt:window top)
			  (vector-ref csr rmd)
			  msk
			  0 0 0
			  65535 65535 65535
			  (cons 1 1)))))
		 500))))
