;;; Test for xmt:ask-for-<something> function

(require 'xmts)
(require 'debug)

;;;;;;;;;;;;;;;;;;;;;;;; Ask For Something ;;;;;;;;;;;;;;;;;;;;;;;

(define (ask-for-string-test widget initstr)
  (let ((s (xmt:ask-for-string widget "AskForString" initstr)))
    (if s 
	(output "Your name is " s ".")
	(output "Sorry, you have no name!"))))

(define (ask-for-string/init widget) (ask-for-string-test widget "Stallone"))
(define (ask-for-string      widget) (ask-for-string-test widget #f))

(define (ask-for-item widget)
  (let ((s (xmt:ask-for-item widget "AskForItem")))
    (if s 
	(output "Your job is " s ".")
	(output "Sorry, nothing selected!"))))

(define (ask-for-item-23 widget must-match)
  (let ((s (xmt:ask-for-item widget "AskForItem2" #("Apple Macintosh" 
						    "Pentium" 
						    "Sun SparcStation")
			     must-match)))
    (if s 
	(output "You have a " s " computer.")
	(output "Sorry, nothing selected!"))))

(define (ask-for-item/dynamic widget) (ask-for-item-23 widget #f))
(define (ask-for-item/match widget) (ask-for-item-23 widget #t))

(define (ask-for-boolean-test widget showcancel)
  (case (xmt:ask-for-boolean widget "AskForBoolean" showcancel)
    ((1) (output "Okay.  Have fun!"))
    ((2) (output "Well, eat later then!"))
    ((3) (output "Dinner has been cancelled."))
    (else (output "oops -- wrong result"))))
(trace ask-for-boolean-test)

(define (ask-for-boolean/cancel widget) (ask-for-boolean-test widget #t))
(define (ask-for-boolean  widget)       (ask-for-boolean-test widget #f))

(define (ask-for-filename widget)
  (let ((L (xmt:ask-for-filename widget "AskForFilename")))
    (if L
	(output "Filename info: " L)
	(output "Filename selection cancelled."))))

;;;;;;;;;;;;;;;;;;;;;;;; Display Messages ;;;;;;;;;;;;;;;;;;;;;;;

(define (display-error widget)
  (xmt:display-message widget "ErrorMessage" xmt:error-message))

(define (display-warning widget)
  (xmt:display-message widget "WarningMessage" xmt:warning-message))

(define (display-information widget)
;  (xmt:display-information-message widget "infomsg"
;				   "Good Morning America!"
;				   "Greeting"
;				   "@fINothing special, just hit OK."))
  (xmt:display-message widget "InformationMessage" xmt:information-message))

;;;;;;;;;;;;;;;;;;;;;;;; Blocking ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define *login-win* '((user   . "username") 
		      (pass   . "password") 
		      (ok     . "OK") 
		      (cancel . "Cancel")
		      (help   . "Help")
		      (layout . "l")))

(define (login widget)
  (if (not (procedure? *login-win*))
      (begin (set! *login-win* 
		   (xmts:build-query-dialog "Login" widget *login-win*))
	     (xmts:add-activate (*login-win* 'ok)     
				(lambda (w) (xmt:unblock 1)))
	     (xmts:add-activate (*login-win* 'cancel) 
				(lambda (w) (xmt:unblock 2)))
	     (xmts:add-activate (*login-win* 'help)   
				(lambda (w) (xmt:display-message 
					     widget "LoginHelp" 
					     xmt:information-message)))))
  (let ((win *login-win*) (user "") (pass ""))
    (xt:manage-children (win 'layout))
    (xm:process-traversal (win 'user) 0)
    (let ((data (xmt:block widget)))
      (case data
	((1) (begin (set! user (xm:text-get-string (win 'user)))
		    (set! pass (xm:text-get-string (win 'pass)))
		    (output "User: " user ", Password: " pass ".")))
	((2) (output "Login Cancelled."))
	(else (output "Oops! Programming Bug!"))))
    (xt:unmanage-children (win 'layout))
    (cons user pass)))

;;;;;;;;;;;;;;;;;;;;;;;; Login ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (new-login w)
  (write (xmt:gui-login w "delilah" "tomkwong" "2testit"))
  (newline))

;;;;;;;;;;;;;;;;;;;;;;;; MsgLine Widget ;;;;;;;;;;;;;;;;;;;;;;;;;

(define (msg-line-set w)
  (let ((m (*main* 'output)))
    (xmt:msg-line-set m "The message is set by xmt:msg-line-set command.")))

(define (msg-line-get w)
  (let ((m (*main* 'output)))
    (xmt:msg-line-set m "How old are you? ")
    (output "You are " (xmt:msg-line-get m xmt:msg-line-int) " years old.")))

(define msg-line-level 0)

(define (msg-line-push w)
  (let ((m (*main* 'output)))
    (set! msg-line-level (+ msg-line-level 1))
	(xmt:msg-line-push m)
	(xmt:msg-line-set m (string-append "This is stack level "
					   (number->string msg-line-level)
					   "."))))

(define (msg-line-pop w)
  (let ((m (*main* 'output)))
    (if (> msg-line-level 0)
	(set! msg-line-level (- msg-line-level 1)))
    (xmt:msg-line-pop m xmt:msg-line-now)))

;;;;;;;;;;;;;;;;;;;;;;;; XmtMenu Widget ;;;;;;;;;;;;;;;;;;;;;;;;;

(define get-menu-item 
  (let ((sensitive #t))
    (lambda (w)
      (let* ((menu (*main* 'askmenu))
	     (item (xmt:menu-get-widget menu "askbool")))
	(if sensitive (set! sensitive #f) (set! sensitive #t))
	(xt:set-sensitive item sensitive)
	(output "Just" (if sensitive "enabled" "disabled") 
		"the widget" (xt:name item) ".")))))

;;;;;;;;;;;;;;;;;;;;;;;; XmtPopupMenu Widget ;;;;;;;;;;;;;;;;;;;;

(define (screen-saver w)
  (let ((name (xt:name w)))
    (output "Starting screen saver with " name "style.")))

(trace screen-saver)

;;;;;;;;;;;;;;;;;;;;;;;; Main Program ;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (user-interrupt) (quit))

;;; This is the information needed to build the toplevel
(define *main* '((btn1 . "askbool") 
		 (btn2 . "askboolcancel") 
		 (btn3 . "askstr")
		 (btn4 . "askstrinit")
		 (btn5 . "askitem")
		 (btn6 . "askitemdynamic")
		 (btn7 . "askitemmatch")
		 (btn8 . "askfile")
		 (btn9 . "displayinfo")
		 (btn10 . "displayerr")
		 (btn11 . "displaywarn")
		 (btn12 . "block")
		 (btn14 . "mlset")
		 (btn15 . "mlget")
		 (btn16 . "mlpush")
		 (btn17 . "mlpop")
		 (btn18 . "getmenuitem")
		 (askmenu . "askmenu")
		 (output . "output")
		 ; screen savers
		 (hop . "hop") (rocks . "rocks")
))

(define (output . args)
  (xmt:msg-line-set
   (*main* 'output) (call-with-output-string 
		     (lambda (port)
		       (do ((L args (cdr L)))
			   ((null? L))
			 (display (car L) port)
			 (display " " port))))))
			
(define (create-main-window top)

  ;(xrm:put-line-resource top "*backgroundPixmap: concrete.xpm")

  ; Returns a closure which show callback data
  (define (show-callback-data str) (lambda (w . data) 
				     (output str "Callback: " data)))

  (set! *main* (xmts:create-query-children top *main*))
  (xmts:add-activate (*main* 'btn1) ask-for-boolean)
  (xmts:add-activate (*main* 'btn2) ask-for-boolean/cancel)
  (xmts:add-activate (*main* 'btn3) ask-for-string)
  (xmts:add-activate (*main* 'btn4) ask-for-string/init)
  (xmts:add-activate (*main* 'btn5) ask-for-item)
  (xmts:add-activate (*main* 'btn6) ask-for-item/dynamic)
  (xmts:add-activate (*main* 'btn7) ask-for-item/match)
  (xmts:add-activate (*main* 'btn8) ask-for-filename)
  (xmts:add-activate (*main* 'btn9) display-information)
  (xmts:add-activate (*main* 'btn10) display-error)
  (xmts:add-activate (*main* 'btn11) display-warning)
  (xmts:add-activate (*main* 'btn12) login)
  (xmts:add-activate (*main* 'btn14) msg-line-set)
  (xmts:add-activate (*main* 'btn15) msg-line-get)
  (xmts:add-activate (*main* 'btn16) msg-line-push)
  (xmts:add-activate (*main* 'btn17) msg-line-pop)
  (xmts:add-activate (*main* 'btn18) get-menu-item)
;  (xmts:add-activate (*main* 'hop) screen-saver)
;  (xmts:add-activate (*main* 'rocks) screen-saver)
  (xmt:register-improved-icons top)
;  (xmt:display-busy-cursor top)

;  (let ((pixmap (xmt:get-pixmap top "concrete.xpm")))
;    (if pixmap
;	(recursive-set-value top xm:n-background-pixmap pixmap)
;	(format #t "Cannot open pixmap file.\n")))

)


(xmts:run-app "combo" create-main-window)


