/*
 * $Header: /home/tomkwong/src/xmtscm/RCS/dt.c,v 1.1 1995/12/05 23:06:07 tomkwong Exp tomkwong $
 *
 * Common Desktop Environment (CDE) widgets extension to SCM.
 * Copyright 1995 Thomas Kwong
 */

#include <Dt/ComboBox.h>
#include <Dt/SpinBox.h>
#include <Dt/MenuButton.h>
#include "scm.h"
#include "x.h"
#include "xt.h"

xt_widget_class_t dt_widget_classes[] = {
    { "dt:combo-box",		&dtComboBoxWidgetClass},
    { "dt:spin-box",		&dtSpinBoxWidgetClass},
    { "dt:menu-button",		&dtMenuButtonWidgetClass},
};

XSymMap dt_symbols[] = {
    /* SpinBox.h */
    {"dt:arrows-flat-beginning", DtARROWS_FLAT_BEGINNING},
    {"dt:arrows-flat-end", DtARROWS_FLAT_END},
    {"dt:arrows-split", DtARROWS_SPLIT},
    {"dt:arrows-beginning", DtARROWS_BEGINNING},
    {"dt:arrows-end", DtARROWS_END},
    {"dt:arrows-sensitive", DtARROWS_SENSITIVE},
    {"dt:arrows-decrement-sensitive", DtARROWS_DECREMENT_SENSITIVE},
    {"dt:arrows-increment-sensitive", DtARROWS_INCREMENT_SENSITIVE},
    {"dt:arrows-insensitive", DtARROWS_INSENSITIVE},
    {"dt:numeric", DtNUMERIC},
    {"dt:string", DtSTRING},
    {"dt:cr-ok", DtCR_OK},
    {"dt:cr-spin-next", DtCR_SPIN_NEXT},
    {"dt:cr-spin-prior", DtCR_SPIN_PRIOR},
    /* ComboBox.h */
    {"dt:left", DtLEFT},
    {"dt:right", DtRIGHT},
    {"dt:motif", DtMOTIF},
    {"dt:windows", DtWINDOWS},
    {"dt:drop-down-list", DtDROP_DOWN_LIST},
    {"dt:drop-down-combo-box", DtDROP_DOWN_COMBO_BOX},
    {"dt:alignment-beginning", DtALIGNMENT_BEGINNING},
    {"dt:alignment-center", DtALIGNMENT_CENTER},
    {"dt:alignment-end", DtALIGNMENT_END},
    {"dt:cr-select", DtCR_SELECT},
    {"dt:cr-menu-post", DtCR_MENU_POST},
    /* MenuButton.h */
    {"dt:cr-cascading", DtCR_CASCADING},
    {0,0}
};

static char s_dt_spin_box_add_item[] = "dt:spin-box-add-item";
SCM dt_spin_box_add_item(w, item, pos)
SCM w, item, pos;
{
  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_dt_spin_box_add_item);
  ASSERT(NIMP(item) && XMSTRINGP(item), item, ARG2, s_dt_spin_box_add_item);
  ASSERT(INUMP(pos), pos, ARG3, s_dt_spin_box_add_item);
  DtSpinBoxAddItem(WIDGET(w), XMSTRING(item), INUM(pos));
  return UNSPECIFIED;
}

static char s_dt_spin_box_delete_pos[] = "dt:spin-box-delete-pos";
SCM  dt_spin_box_delete_pos(w, pos)
SCM w, pos;
{
  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_dt_spin_box_delete_pos);
  ASSERT(INUMP(pos), pos, ARG2, s_dt_spin_box_delete_pos);
  DtSpinBoxDeletePos(WIDGET(w), INUM(pos));
  return UNSPECIFIED;
}

static char s_dt_spin_box_set_item[] = "dt:spin-box-set-item";
SCM  dt_spin_box_set_item(w, item)
SCM w, item;
{
  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_dt_spin_box_set_item);
  ASSERT(NIMP(item) && XMSTRINGP(item), item, ARG2, s_dt_spin_box_set_item);
  DtSpinBoxSetItem(WIDGET(w), XMSTRING(item));
  return UNSPECIFIED;
}

static char s_dt_combo_box_add_item[] = "dt:combo-box-add-item";
SCM dt_combo_box_add_item(w, item, args)
SCM w, item, args;
{
  SCM pos, unique;

  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_dt_combo_box_add_item);
  ASSERT(NIMP(item) && XMSTRINGP(item), item, ARG2, s_dt_combo_box_add_item);
  ASRTARG(pos, args, ARG3, s_dt_combo_box_add_item);
  ASSERT(INUMP(pos), pos, ARG3, s_dt_combo_box_add_item);
  ASRTARG(unique, args, ARG4, s_dt_combo_box_add_item);
  ASSERT(BOOLP(unique), unique, ARG4, s_dt_combo_box_add_item);

  DtComboBoxAddItem(WIDGET(w), XMSTRING(item), INUM(pos), BOOLVAL(unique));
  return UNSPECIFIED;
}

static char s_dt_combo_box_delete_pos[] = "dt:combo-box-delete-pos";
SCM  dt_combo_box_delete_pos(w, pos)
SCM w, pos;
{
  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_dt_combo_box_delete_pos);
  ASSERT(INUMP(pos), pos, ARG2, s_dt_combo_box_delete_pos);
  DtComboBoxDeletePos(WIDGET(w), INUM(pos));
  return UNSPECIFIED;
}

static char s_dt_combo_box_set_item[] = "dt:combo-box-set-item";
SCM  dt_combo_box_set_item(w, item)
SCM w, item;
{
  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_dt_combo_box_set_item);
  ASSERT(NIMP(item) && XMSTRINGP(item), item, ARG2, s_dt_combo_box_set_item);
  DtComboBoxSetItem(WIDGET(w), XMSTRING(item));
  return UNSPECIFIED;
}

static char s_dt_combo_box_select_item[] = "dt:combo-box-select-item";
SCM  dt_combo_box_select_item(w, item)
SCM w, item;
{
  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_dt_combo_box_select_item);
  ASSERT(NIMP(item) && XMSTRINGP(item), item, ARG2,
	 s_dt_combo_box_select_item);
  DtComboBoxSelectItem(WIDGET(w), XMSTRING(item));
  return UNSPECIFIED;
}

iproc cde_subr2s[] = {
  {s_dt_spin_box_delete_pos,    dt_spin_box_delete_pos},
  {s_dt_spin_box_set_item,      dt_spin_box_set_item},
  {s_dt_combo_box_delete_pos,   dt_combo_box_delete_pos},
  {s_dt_combo_box_set_item,     dt_combo_box_set_item},
  {s_dt_combo_box_select_item,	dt_combo_box_select_item},
  {0,0},
};

iproc cde_subr3s[] = {
  {s_dt_spin_box_add_item,	dt_spin_box_add_item},
  {0,0},
};

iproc cde_lsubr2s[] = {
  {s_dt_combo_box_add_item,	dt_combo_box_add_item},
  {0, 0}
};

void init_cde()
{
  x_make_integer_constants(dt_symbols);
  init_iprocs(cde_subr2s,  tc7_subr_2);
  init_iprocs(cde_subr3s,  tc7_subr_3);
  init_iprocs(cde_lsubr2s, tc7_lsubr_2);
  xt_init_widget_classes(
      dt_widget_classes,
      XtNumber(dt_widget_classes),
      "*cde-motif-widget-classes*");
#ifdef XMT
  XmtRegisterWidgetClass("DtComboBox",   dtComboBoxWidgetClass);
  XmtRegisterWidgetClass("DtSpinBox",    dtSpinBoxWidgetClass);
  XmtRegisterWidgetClass("DtMenuButton", dtMenuButtonWidgetClass);
#endif
}
