(require 'xm)
(require 'xmconst)
(require 'common-list-functions)

(define (run name class proc)
  (let ((toplevel (xt:app-initialize name class)))
    (proc toplevel)
    (xt:realize-widget toplevel)
    (xt:app-main-loop toplevel)))

(define (create-grid parent name rows types)
  (let* ((scrolled-window (xt:create-widget name xm:scrolled-window parent))
	 (column          (xt:create-managed-widget "rc" xm:row-column
						    scrolled-window))
	 (columns         (length types)))
    (do ((i 0 (1+ i)))
	((= i rows))
      (let ((row (xt:create-managed-widget
		  (string-append "row_" (number->string i))
		  xm:row-column
		  column
		  xm:n-orientation xm:horizontal)))
	(for-each (lambda (type)
		    (apply xt:create-managed-widget
			   (append (list (car type) (cadr type) row)
				   (cddr type))))
		  types)))
    (xt:manage-children scrolled-window)
    column))

(define (grid:get-widget grid row column)
  (let* ((nrows (xt:get-value grid xm:n-num-children xt:integer))
	 (first (xt:name->widget grid "row_1"))
	 (ncols (xt:get-value first xm:n-num-children xt:integer)))
    (write nrows) (newline) (write ncols) (newline)
    (if (and (< 0 row (1+ nrows)) (< 0 column (1+ ncols)))
	(let* ((name   (string-append "row_" (number->string row)))
	       (dstrow (xt:name->widget grid name)))
	  (if dstrow
	      (let ((wlist (xt:get-value dstrow xm:n-children xt:widgetlist
					 ncols)))
		(vector-ref wlist (1- column)))
	      #f))
	(error 'invalid-row-column-range row column))))

(define grid-columns '(10 5 20))

(define grid-settings (map (lambda (col) (list "abc" xm:text xm:n-columns col))
			   grid-columns))

;(write grid-settings)(newline)

(run "grid" "Grid"
     (lambda (top)
       (for-each (lambda (line) (xrm:put-line-resource top line))
		 '("*XmRowColumn.marginWidth: 0"
		   "*XmRowColumn.marginHeight: 0"
		   "*XmRowColumn.spacing: 0"))
       (let ((g (create-grid top "g" 10 grid-settings)))
	 (write (grid:get-widget g 2 3)) (newline))))
