;;; Testing callback data.

(require 'xmts)
(require 'format)
(load "bubble")
(load "password")

(xmts:run-app

 "Newcombo"

 (lambda (top)

   (set! *xmt-debug-mode* #t)

   (let ((w (xmts:create-query-children top '((icons . "icons")
					      (list . "list")
					      (size . "fontsize")
					      (text . "text")
					      (push1 . "p1")
					      (push2 . "p2")
					      (arrow . "arrow")
					      (cartoon . "cartoon")
					      (scale . "scale")
					      (bar   . "bar")
					      ))))

     ; List
     (xmts:list-set-items-ascii (w 'list)
				#("Thomas" "Eric" "Alex" "Billy" "Joe"))
     (xmts:add-extended (w 'list) 
			(lambda (w c d)
			  (debug d))
			#f)

     ; Chooser
     (xt:add-callback (w 'size) xm:n-value-changed-callback
		      (lambda (w c d)
			(format #t "~X, " (xmt:get-chooser-state w))
			(format #t "~D"   (xmt:chooser-get-value w xt:integer))
			(newline))
		      #f)

     ; Text
     (let ((passobj (xmts:make-password-text (w 'text))))
       (xmts:add-activate (w 'text)
			  (lambda (w)
			    (display "Oh your password is: ")
			    (display (passobj))
			    (newline))))

     ; Push Button
     (xmts:add-activate (w 'push1)
			(lambda (w c d)
			  (debug "push button:" c ",data:" d))
			1)
     (xmts:add-activate (w 'push2)
			(lambda (w c d)
			  (debug "push button:" c ",data:" d))
			2)

     ; Arrow Button
     (xmts:add-activate
      (w 'arrow)
      (let* ((names (list "batsignal"
			  "bill_cat"
			  "bsd_daemon"
			  "bsd_daemon2"
			  "cat_scratch" 
			  "caution"
			  "city"
			  "clipboard" 
			  "earth"
			  "fireplace"
			  "flowers" 
			  "gary"
			  "iris"
			  "plane"
			  "robotman"
			  "spacecop"
			  "bart"
			  "hobbes"))
	     (files (map (lambda (s) (string-append "icons/" s ".xpm"))
			 names))
	     (count 0)
	     (max   (- (length files) 1))
	     (lastpix #f))
	(lambda (widget)
	  (let ((pixmap (xmt:get-pixmap widget (list-ref files count))))
	    (if pixmap
		(begin (if lastpix (xmt:release-pixmap widget lastpix))
		       (xt:set-values (w 'cartoon)
				      xm:n-label-pixmap pixmap)
		       (set! lastpix pixmap))
		(debug "Cannot find cartoon file."))
	    (set! count (if (< count max) (+ count 1) 0))))))

     ; Scale
     (xmts:add-drag (w 'scale)
		    (lambda (w c d)
			  (debug "scale:" (vector-ref d 2)))
		    #f)

     ; Scroll Bar
     (xmts:add-drag (w 'bar)
		    (lambda (w c d)
			  (debug "bar:" (vector-ref d 2)))
		    #f)

     ; Icon bubble help
     (for-each (lambda (s)
		 (xmts:add-bubble-help 
		  (xmt:name-to-widget (w 'icons) (string-append "." s))
		  (let ((s s))
		    (string-set! s 0 (char-upcase (string-ref s 0)))
		    s)
		  1000))
	       '("afghanis" "bahrain" "banglade" "belgium" "bolivia"
	         "canada" "chile" "china" "congo" "denmark"))
					  
)))

