/* Password Text Entry Callbacks */
#include <string.h>
#include <stdlib.h>
#include <memory.h>
#include <Xm/Text.h>
#include <Xmt/Xmt.h>
#include <Xmt/Procedures.h>

#define MAX_PASSWORD_LEN 24

static Boolean dont_check = False;

void
CB_PasswordTextVerify(Widget w, XtPointer p, XtPointer q)
{
    char *word;
    int len;
    XmTextVerifyCallbackStruct *cbs = (XmTextVerifyCallbackStruct*)q;

    if (dont_check)
	return;

    /* Uses XmNuserData to store password */
    XtVaGetValues(w, XmNuserData, &word, NULL);
    if (!word) {
	word = (char*) calloc(MAX_PASSWORD_LEN + 1, sizeof(char));
	XtVaSetValues(w, XmNuserData, (XtPointer)word, NULL);
    }

    if (cbs->startPos < cbs->currInsert) { /* backspace */
	cbs->endPos = strlen(word);
	word[cbs->startPos] = 0;
	return;
    }
	
    if (cbs->text->length > 1) {	/* no cut and paste */
	cbs->doit = False;
	return;
    }

    if (strlen(word) == MAX_PASSWORD_LEN) {
	cbs->doit = False;
	return;
    }

    /* now we can assume that there is only one character in cbs->text->ptr */
    len = strlen(word);
    word[len] = *(cbs->text->ptr);	/* append only one char */
    word[len+1] = 0;

    for (len = 0; len < cbs->text->length; len++)
	cbs->text->ptr[len] = '*';
}

void
CB_PasswordTextMotion(Widget w, XtPointer p, XtPointer data)
{
    XmTextVerifyCallbackStruct *cbs = (XmTextVerifyCallbackStruct *)data;

    if (cbs->newInsert != XmTextGetLastPosition(w))
	cbs->doit = False;
}

String
PasswordTextGetString(Widget w)
{
    char *word;

    XtVaGetValues(w, XmNuserData, &word, NULL);
    if (word) 
	return XtNewString(word);
    else
	return XtNewString("");
}

void
PasswordTextSetString(Widget w, String str)
{
    char *word = XtNewString(str);
    char *word2 = XtNewString(str);
    int i;

    for (i = 0; i < strlen(word2); i++)
	word2[i] = '*';

    dont_check = True;
    XtVaSetValues(w, XmNuserData, word, XmNvalue, word2, NULL);
    dont_check = False;

    XmTextSetInsertionPosition(w, XmTextGetLastPosition(w));

    XtFree(word2);
}

void
PasswordTextClearString(Widget w)
{
    char *word;
    int i;

    XtVaGetValues(w, XmNuserData, &word, NULL);
    if (word) {
	for (i = 0; i < MAX_PASSWORD_LEN; i++)
	    word[i] = 0;
    }
}

void
PasswordTextRegisterXmt()
{
    static XmtProcedureInfo procs[] = {
	{ 
	    "CB_PasswordTextVerify", (XmtProcedure)CB_PasswordTextVerify,
	    {XmtRCallbackWidget, XmtRCallbackUnused, XmtRCallbackData, 
		 0,0,0,0,0}
	},
	{
	    "CB_PasswordTextMotion", (XmtProcedure)CB_PasswordTextMotion,
	    {XmtRCallbackWidget, XmtRCallbackUnused, XmtRCallbackData, 
		 0,0,0,0,0 }
	},
    };
    XmtRegisterProcedures(procs, XtNumber(procs));
}


