;;; Rheostat (A drawing area)

(require 'xmts)

(define (create-rheostat parent name value min max value->string)

  (let* ((dpy   (xt:display parent))
	 (frame (xt:create-managed-widget "frame" xm:frame parent))
	 (label (xt:create-managed-widget "value" xm:label frame))
	 (gc   (x:create-gc dpy '() x:gc-foreground 1 x:gc-background 0))
	 (cmap  (x:default-colormap dpy 0))
	 (cell  (x:alloc-color dpy cmap 25600 25600 65535)))

    (define (set-value new-value)
      (if (xt:is-realized label)
	  (let* ((width  (xt:get-value label xm:n-width  xt:integer))
		 (height (xt:get-value label xm:n-height xt:integer))
		 (win  (xt:window label)))
	    (set! value new-value)
	    (xmts:label-set-string label (value->string value))
	    (let ((width (quotient (* width value) max)))
	      (x:fill-rectangle dpy win gc 0 0 width height)))
	  (xmts:program-warning "Cannot set rheostat value because the widget is not realized yet." label)))

    (define (init widget event)
      (set-value value)
      (xt:remove-event-handler label x:exposure-mask 0 init))

    (x:set-foreground dpy gc cell)
    (xt:add-event-handler label x:exposure-mask 0 init)
    (lambda (msg . args)
      (case msg
	((set)   (apply set-value args))
	((clear) (apply set-value '(0)))
	(else  (error "Unknown rheostat command." msg))))))

(xmts:run-app 
 "rheostat" 
 (lambda (top)
   (let* ((lay (xt:create-managed-widget "lay" xmt:layout top))
	  (rh  (create-rheostat lay "rh" 0 0 100
				(lambda (v) (string-append 
					     (number->string v)
					     "%"))))
	  (bar (xt:create-managed-widget "scale" xm:scale lay
					 xm:n-minimum 0
					 xm:n-maximum 100
					 xm:n-orientation xm:horizontal
					 xmt:n-layout-stretchability 0))
;	  (up   (xt:create-managed-widget "up" xm:arrow-button lay
;					  xm:n-arrow-direction xm:arrow-up))
;	  (down (xt:create-managed-widget "down" xm:arrow-button lay
;					  xm:n-arrow-direction xm:arrow-down))
	  (curr 0)
	  (curr+ (lambda ()
		   (if (< curr 100)
		       (set! curr (+ curr 10)))
		   curr))
	  (curr- (lambda ()
		   (if (> curr 0)
		       (set! curr (- curr 10)))
		   curr))
	  (act (lambda (widget direction)
		 (case direction
		   ((up)   (rh 'set (curr+)))
		   ((down) (rh 'set (curr-))))))
	  )
;     (xmts:add-activate up   act 'up)
;     (xmts:add-activate down act 'down)
     (xmts:add-drag bar (lambda (widget c d)
			  (rh 'set (xm:scale:value d)))
		    #f)
			  
     
)))

