/*
 * $Header: /home/tomkwong/src/xmtscm/RCS/xmt.c,v 1.2 1995/12/06 23:10:05 tomkwong Exp $
 *
 * Xmt Library Extension for SCM.
 * Copyright 1995 by Thomas Kwong (tomkwong@wilshire.com)
 * 
 */

#include <Xmt/All.h>
#include "scm.h"
#include "x.h"
#include "xt.h"

extern void protect_callback(SCM);
extern SCM  make_xpixmap();

/* New scheme objects */

#define XMTIMAGEP(x) (TYP16(x) == tc16_xmtimage)
#define XMTIMAGE(x) ((XmtImage*) CDR(x))
long tc16_xmtimage;
static sizet free_xmtimage();
static int   print_xmtimage();
static SCM   equalp_xmtimage();

static char s_xmt_initialize[]		    = "xmt:initialize";
static char s_xmt_create_children[]	    = "xmt:create-children";
static char s_xmt_name_to_widget[]	    = "xmt:name-to-widget";
static char s_xmt_ask_for_string[]     	    = "xmt:ask-for-string";
static char s_xmt_ask_for_item[]     	    = "xmt:ask-for-item";
static char s_xmt_ask_for_boolean[]         = "xmt:ask-for-boolean";
static char s_xmt_ask_for_filename[]        = "xmt:ask-for-filename";
static char s_xmt_display_message[]         = "xmt:display-message";
static char s_xmt_display_busy_cursor[]     = "xmt:x-display-busy/default-cursor";
static char s_xmt_block[]                   = "xmt:block";
static char s_xmt_unblock[]                 = "xmt:unblock";
static char s_xmt_generic_get_shell[]       = "xmt:x-get-shell";
static char s_xmt_create_or_change_pixmap_icon[] = "xmt:x-create/change-pixmap-icon";
static char s_xmt_wait_until_mapped[]       = "xmt:wait-until-mapped";
static char s_xmt_register_improved_icons[] = "xmt:register-improved-icons";
static char s_xmt_chooser_get_value[]       = "xmt:chooser-get-value";
static char s_xmt_msg_line_set[]            = "xmt:msg-line-set";
static char s_xmt_msg_line_clear[]          = "xmt:msg-line-clear";
static char s_xmt_msg_line_append[]         = "xmt:msg-line-append";
static char s_xmt_msg_line_push[]           = "xmt:msg-line-push";
static char s_xmt_msg_line_pop[]            = "xmt:msg-line-pop";
static char s_xmt_msg_line_get[]            = "xmt:msg-line-get";
static char s_xmt_menu_get_widget[]         = "xmt:menu-get-widget";
static char s_xmt_set_application_values[]  = "xmt:set-application-values";
static char s_xmt_get_bitpixmap[]           = "xmt:x-get-bitmap/pixmap";
static char s_xmt_release_pixmap[]          = "xmt:release-pixmap";
static char s_xmt_manipulate_shell[]        = "xmt:x-manipulate-shell";
static char s_xmt_dialog_position[]         = "xmt:dialog-position";
static char s_xmt_add_wm_callback[]         = "xmt:x-add-wm-callback";
static char s_xmt_layout_xable[]            = "xmt:x-layout-enable/disable";
static char s_xmt_parse_xpm_file[]          = "xmt:parse-xpm-file";
static char s_xmt_get_xmt_image_field[]     = "xmt:get-xmt-image-field";
static char s_xmt_create_pixmap_from_xmt_image[] = "xmt:create-pixmap-from-xmt-image";
static char s_xmt_load_resource_file[] = "xmt:load-resource-file";

#ifdef USE_RESOURCE_MANAGER
static char s_xrm_put_line_resource[]       = "xrm:put-line-resource";
#endif

static XSymMap xmt_symbols[] = {
  {"xmt:chooser-radio-box", XmtChooserRadioBox},
  {"xmt:chooser-check-box", XmtChooserCheckBox},
  {"xmt:chooser-radio-palette", XmtChooserRadioPalette},
  {"xmt:chooser-check-palette", XmtChooserCheckPalette},
  {"xmt:chooser-radio-list", XmtChooserRadioList},
  {"xmt:chooser-check-list", XmtChooserCheckList},
  {"xmt:chooser-option", XmtChooserOption},
  {"xmt:chooser-button-box", XmtChooserButtonBox},
  {"xmt:cancel-button", XmtCancelButton},
  {"xmt:ok-button", XmtOkButton},
  {"xmt:help-button", XmtHelpButton},
  {"xmt:apply-button", XmtApplyButton},
  {"xmt:reset-button", XmtResetButton},
  {"xmt:done-button", XmtDoneButton},
  {"xmt:stop-button", XmtStopButton},
  {"xmt:yes-button", XmtYesButton},
  {"xmt:no-button", XmtNoButton},
  {"xmt:help-format-string", XmtHelpFormatString},
  {"xmt:help-format-xm-string", XmtHelpFormatXmString},
  {"xmt:layout-filled", XmtLayoutFilled},
  {"xmt:layout-flush-left", XmtLayoutFlushLeft},
  {"xmt:layout-flush-top", XmtLayoutFlushTop},
  {"xmt:layout-flush-right", XmtLayoutFlushRight},
  {"xmt:layout-flush-bottom", XmtLayoutFlushBottom},
  {"xmt:layout-centered", XmtLayoutCentered},
  {"xmt:layout-top", XmtLayoutTop},
  {"xmt:layout-bottom", XmtLayoutBottom},
  {"xmt:layout-left", XmtLayoutLeft},
  {"xmt:layout-right", XmtLayoutRight},
  {"xmt:layout-frame-none", XmtLayoutFrameNone},
  {"xmt:layout-frame-box", XmtLayoutFrameBox},
  {"xmt:layout-frame-left", XmtLayoutFrameLeft},
  {"xmt:layout-frame-right", XmtLayoutFrameRight},
  {"xmt:layout-frame-top", XmtLayoutFrameTop},
  {"xmt:layout-frame-bottom", XmtLayoutFrameBottom},
  {"xmt:layout-frame-shadow-in", XmtLayoutFrameShadowIn},
  {"xmt:layout-frame-shadow-out", XmtLayoutFrameShadowOut},
  {"xmt:layout-frame-etched-in", XmtLayoutFrameEtchedIn},
  {"xmt:layout-frame-etched-out", XmtLayoutFrameEtchedOut},
  {"xmt:layout-frame-single-line", XmtLayoutFrameSingleLine},
  {"xmt:layout-frame-double-line", XmtLayoutFrameDoubleLine},
  {"xmt:layout-frame-inside", XmtLayoutFrameInside},
  {"xmt:layout-frame-through", XmtLayoutFrameThrough},
  {"xmt:layout-frame-outside", XmtLayoutFrameOutside},
  {"xmt:layout-space-none", XmtLayoutSpaceNone},
  {"xmt:layout-space-even", XmtLayoutSpaceEven},
  {"xmt:layout-space-l-r-even", XmtLayoutSpaceLREven},
  {"xmt:layout-space-interval", XmtLayoutSpaceInterval},
  {"xmt:layout-space-l-c-r", XmtLayoutSpaceLCR},
  {"xmt:layout-space-l-tabbed", XmtLayoutSpaceLTabbed},
  {"xmt:layout-space-c-tabbed", XmtLayoutSpaceCTabbed},
  {"xmt:layout-space-r-tabbed", XmtLayoutSpaceRTabbed},
  {"xmt:layout-points", XmtLayoutPoints},
  {"xmt:layout-inches", XmtLayoutInches},
  {"xmt:layout-millimeters", XmtLayoutMillimeters},
  {"xmt:layout-ems", XmtLayoutEms},
  {"xmt:layout-ens", XmtLayoutEns},
  {"xmt:menu-item-end", XmtMenuItemEnd},
  {"xmt:menu-item-push-button", XmtMenuItemPushButton},
  {"xmt:menu-item-toggle-button", XmtMenuItemToggleButton},
  {"xmt:menu-item-cascade-button", XmtMenuItemCascadeButton},
  {"xmt:menu-item-separator", XmtMenuItemSeparator},
  {"xmt:menu-item-double-separator", XmtMenuItemDoubleSeparator},
  {"xmt:menu-item-label", XmtMenuItemLabel},
  {"xmt:menu-item-on", XmtMenuItemOn},
  {"xmt:menu-item-help", XmtMenuItemHelp},
  {"xmt:menu-item-tearoff", XmtMenuItemTearoff},
  {"xmt:menu-item-pixmap", XmtMenuItemPixmap},
  {"xmt:menu-item-callback-list", XmtMenuItemCallbackList},
/* These are used as booleans in xmtscm */
/*   {"xmt:msg-line-now", XmtMsgLineNow}, */
/*   {"xmt:msg-line-on-action", XmtMsgLineOnAction}, */
  {"xmt:search-path-only", XmtSearchPathOnly},
  {"xmt:search-user-path", XmtSearchUserPath},
  {"xmt:search-app-path", XmtSearchAppPath},
  {"xmt:search-sys-path", XmtSearchSysPath},
  {"xmt:search-everywhere", XmtSearchEverywhere},
  {"xmt:version", XmtVersion},
  {"xmt:patchlevel", XmtPatchlevel},
  {0,0}
};

/* These variables are used to support XmtBlock() */
static Boolean blocking;
static SCM *block_result;

xt_widget_class_t xmt_widget_classes[] = {
  {"xmt:chooser",              &xmtChooserWidgetClass},
  {"xmt:cli",                  &xmtCliWidgetClass},
  {"xmt:help-box",             &xmtHelpBoxWidgetClass},
  {"xmt:input-field",          &xmtInputFieldWidgetClass},
  {"xmt:layout",		&xmtLayoutWidgetClass},
  {"xmt:layout-box",           &xmtLayoutBoxGadgetClass},
  {"xmt:layout-pixmap",        &xmtLayoutPixmapGadgetClass},
  {"xmt:layout-separator",     &xmtLayoutSeparatorGadgetClass},
  {"xmt:layout-space",         &xmtLayoutSpaceGadgetClass},
  {"xmt:layout-string",        &xmtLayoutStringGadgetClass},
  {"xmt:menu",                 &xmtMenuWidgetClass},
  {"xmt:msg-line",             &xmtMsgLineWidgetClass},
  {"xmt:working-box",          &xmtWorkingBoxWidgetClass},
};

static smobfuns xmtimage_smob = {
  mark_no_further, 
  free_xmtimage,			/* why free_mailstream not working? */
  print_xmtimage,
  equalp_xmtimage
};

SCM make_xmtimage()
{
  SCM s;
  NEWCELL(s);
  CAR(s) = tc16_xmtimage;
  CDR(s) = 0;
  return s;
}

sizet free_xmtimage(ptr)
SCM ptr;
{
  XmtFreeXmtImage(XMTIMAGE(ptr));
  return sizeof(XmtImage);
}

int print_xmtimage(exp, f, writing)
SCM exp;
SCM f;
int writing;
{
  lputs("#<XmtImage>", f);
  return 1;
}

SCM equalp_xmtimage(s1, s2)
SCM s1, s2;
{
  return VALBOOL(XMTIMAGE(s1) == XMTIMAGE(s2));
}

/* This procedure may be called as an action specified in
 * a resource file.  The parameters should be valid SCM
 * codes(which may call some global scheme procedures).
 */
static void scm_action(w, event, params, num)
Widget w;
XEvent *event;
String *params;
Cardinal *num;
{
  Cardinal i;
  int len;
  char *str;
  SCM result;

  for (i = 0; i < *num; i++)
  {
    len = 3 + strlen(params[i]); /* including parentheses and null */
    str = (char*) must_malloc(sizeof(char) * len, "scm_action");
    sprintf(str, "(%s)", params[i]);
    result = scm_evstr(str);
/*    iprin1(result, cur_output_port(), 1);*/
    must_free(str);
  }
}

static XtActionsRec scmactions[] = {
  { "scm", scm_action },
};


/* This procedure may be called from any callback resources as
 * specified in the resource file.  It is registered as a Xmt
 * proceudre accepting only one argument which must be a valid
 * scheme code.
 */
static void scm_procedure(str)
String str;
{
  char *s = must_malloc(sizeof(char) * (strlen(str)+3), "scm_procedure");
  sprintf(s, "(%s)", str);
  scm_evstr(s);
  must_free(s);
}

static XmtProcedureInfo xmtprocs[] = {
  { "scm", (XmtProcedure)scm_procedure, {XtRString,0,0,0,0,0,0} },
};


/* This is used in placed of XmtRegisterAll because I don't
 * want to register the string to font list converter in CDE
 * environment.
 */
static void xmt_register_stuff()
{
    XmtRegisterMotifWidgets();
    XmtRegisterXmtWidgets();

    XmtRegisterUnixProcedures();
    XmtRegisterXtProcedures();
    XmtRegisterXmtProcedures();

    /* these are also registered by XmtInitialize() */
    XmtRegisterBitmapConverter();
    XmtRegisterBitmaskConverter();
    XmtRegisterPixmapConverter();
    XmtRegisterColorTableConverter();
    XmtRegisterWidgetConverter();
    XmtRegisterCallbackConverter();
    XmtRegisterXmStringConverter(); 
#ifndef CDE
    XmtRegisterXmFontListConverter();
#endif
    XmtRegisterStringListConverter();
    XmtRegisterMenuItemsConverter();
    XmtRegisterPixmapListConverter();
    XmtRegisterPixelConverter();

    XmtRegisterLayoutParser();
    XmtRegisterLayoutCreateMethod();
}

SCM xmt_initialize(sname, sclass)
SCM sname, sclass;
{
  XtAppContext app;
  Widget top_level;
  char **argv, *name;
  int argc;
#ifdef DEBUG_XMT
  int i;
#endif
  extern void _XEditResCheckMessages();

  ASSERT(NIMP(sname) && STRINGP(sname), sname, ARG1, s_xmt_initialize);
  ASSERT(NIMP(sclass) && STRINGP(sclass), sclass, ARG2, s_xmt_initialize);

  /* Make argv from program arguments.  We will restore it later. */
  argv = makargvfrmstrs(CDR(progargs), s_xmt_initialize);
  for (argc = 0; argv[argc] != 0; argc++) 
    ;

  DEFER_INTS;

  /* Use user-supplied name instead of command name. */
  name = argv[0];
  argv[0] = CHARS(sname);

#ifdef DEBUG_XMT
  for (i = 0; i < argc; i++)
    printf("BEFORE: argv[%d]=%s\n", i, argv[i]);
#endif

  top_level = XtAppInitialize(&app, CHARS(sclass), NULL, 0, 
			      &argc, argv,
			      NULL, NULL, 0);
  ASSERT(top_level != 0, sclass, "XmtInitialize error", s_xmt_initialize);

  XmtParseCommandLine(top_level, &argc, argv);
  XmtRegisterLayoutParser();
  XmtRegisterLayoutCreateMethod();
  XmtInitializeApplicationShell(top_level, NULL, 0);

#ifdef DEBUG_XMT
  for (i = 0; i < argc; i++)
    printf("AFTER: argv[%d]=%s\n", i, argv[i]);
#endif

  /* Restore program arguments */
  argv[0] = name;			/* restore program name */
  progargs = EOL;			/* XXX - do I need this? */
  progargs = makfromstrs(argc, argv);

  ALLOW_INTS;

#ifdef DEBUG_XMT
  { 
    String name, class;
    XtGetApplicationNameAndClass(XtDisplay(top_level), &name, &class);
    printf("name=[%s] class=[%s]\n", name, class);
  }
#endif

  xmt_register_stuff();
  XmtRegisterProcedures(xmtprocs, XtNumber(xmtprocs));
  XtAppAddActions(app, scmactions, XtNumber(scmactions));
  XtAddEventHandler(top_level, (EventMask) 0, True,
		    _XEditResCheckMessages, NULL);

  return make_widget(top_level);
}

SCM xmt_create_children(parent)
SCM parent;
{
  XmtCreateChildren(WIDGET(parent));
  return UNSPECIFIED;
}

SCM xmt_name_to_widget(ref, name)
SCM ref, name;
{
  Widget w;
  ASSERT(NIMP(ref) && WIDGETP(ref), ref, ARG1, s_xmt_name_to_widget);
  ASSERT(NIMP(name) && STRINGP(name), name, ARG2, s_xmt_name_to_widget);
  w = XmtNameToWidget(WIDGET(ref), CHARS(name));
  return (w != NULL) ? make_widget(w) : BOOL_F;
}

SCM xmt_generic_get_shell(w, kind)
SCM w, kind;
{
  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_generic_get_shell);
  ASSERT(INUMP(kind) && INUM(kind) >= 1 && INUM(kind) <= 3, kind,
	 ARG2, s_xmt_generic_get_shell);
  switch (INUM(kind)) {
  case 1:
    return make_widget(XmtGetShell(WIDGET(w)));
  case 2:
    return make_widget(XmtGetApplicationShell(WIDGET(w)));
  case 3:
    return make_widget(XmtGetTopLevelShell(WIDGET(w)));
  }
  return BOOL_F;			/* unreachable */
}

/* Returns 1 for yes, 2 for no, 3 for cancel. */
SCM xmt_ask_for_boolean(w, query_name, args)
SCM w, query_name, args;
{
  SCM showcancel = BOOL_F;
  Boolean b = False;

  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_ask_for_boolean);
  ASSERT(NIMP(query_name) && STRINGP(query_name), query_name, ARG2, 
	 s_xmt_ask_for_boolean);
  
  GETARG(showcancel, args);
  ASSERT(BOOLP(showcancel), showcancel, ARG3, s_xmt_ask_for_string);

  if (XmtAskForBoolean(WIDGET(w), CHARS(query_name), NULL,
		       NULL, NULL, NULL, XmtCancelButton, XmDIALOG_QUESTION, 
		       BOOLVAL(showcancel), &b, NULL))
  {
    return b ? MAKINUM(1) : MAKINUM(2);
  }
  return MAKINUM(3);
}

/* Return a SCM string if user enterred and pressed OK, #f otherwise */
SCM xmt_ask_for_string(w, name, init)
SCM w, name, init;
{
  char buf[1024];

  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_ask_for_string);
  ASSERT(NIMP(name) && STRINGP(name), name, ARG2, s_xmt_ask_for_string);
  ASSERT(FALSEP(init) || (NIMP(init) && STRINGP(init)), init, ARG3, 
	 s_xmt_ask_for_string);
  
  if (!FALSEP(init)) {
    strncpy(buf, CHARS(init), 1023);
    buf[1023] = '\0';
  } else {
    buf[0] = '\0';
  }
  if (XmtAskForString(WIDGET(w), CHARS(name), NULL, buf, 1024, NULL))
    return makfrom0str(buf);
  else
    return BOOL_F;
}

SCM xmt_ask_for_item(w, name, args)
SCM w, name, args;
{
  SCM items, match;
  char buf[1024];
  String *table = NULL;
  unsigned long i, len;

  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_ask_for_item);
  ASSERT(NIMP(name) && STRINGP(name), name, ARG2, s_xmt_ask_for_item);
  GETARG(items, args);
  GETARG(match, args);

  if (!FALSEP(items)) {
    ASSERT(NIMP(items) && VECTORP(items), items, ARG3, s_xmt_ask_for_item);
    table = (String*) XtMalloc(sizeof(String) * (len = LENGTH(items)));
    for (i = 0; i < len; i++) {
      table[i] = XtNewString(CHARS(VELTS(items)[i]));
    }
  }
  ASSERT(BOOLP(match), match, ARG4, s_xmt_ask_for_item);

  buf[0] = '\0';
  if (XmtAskForItem(WIDGET(w), CHARS(name), NULL, NULL, 
		    BOOLCVT(items, table, NULL),
		    BOOLCVT(items, len, 0),
		    BOOLVAL(match),
/* 		    FALSEP(items) ? NULL : table,  */
/* 		    FALSEP(items) ? 0    : len, */
/* 		    FALSEP(match) ? False : True,  */
		    buf, 1024, NULL))
  {
    if (!FALSEP(items)) {
      for (i = 0; i < len; i++)
	XtFree(table[i]);
      XtFree((void*)table);
    }
    return makfrom0str(buf);
  }
  return BOOL_F;
}

SCM xmt_ask_for_filename(w, name, args)
SCM w, name, args;
{
  char fname[1024], dir[1024], pat[1024];

  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_ask_for_filename);
  ASSERT(NIMP(name) && STRINGP(name), name, ARG2, s_xmt_ask_for_filename);
  fname[0] = dir[0] = pat[0] = '\0';
  if (NIMP(args) && CONSP(args)) {
    SCM sdir = CAR(args);
    ASSERT(NIMP(sdir) && STRINGP(sdir), sdir, ARG3, s_xmt_ask_for_filename);
    strcpy(dir, CHARS(sdir));
  }
  if (XmtAskForFilename(WIDGET(w), CHARS(name), NULL, NULL, NULL,
			fname, 1024, dir, 1024, pat, 1024, NULL))
    return cons(makfrom0str(fname), 
		cons(makfrom0str(dir), 
		     cons(makfrom0str(pat), EOL)));
  else
    return BOOL_F;
}

SCM xmt_display_message(w, name, args)
SCM w, name, args;
{
  SCM type, msg, title, help;

  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_display_message);
  ASSERT(NIMP(name) && STRINGP(name), name, ARG2, s_xmt_display_message);
  ASSERT(!NULLP(args), args, ARG3, s_xmt_display_message);
  type = CAR(args); args = CDR(args);
  ASSERT(INUM(type) > 0 && INUM(type) < 4, type, ARG3, s_xmt_display_message);
  GETARG(msg, args);
  ASSERT(msg == BOOL_F || (NIMP(msg) && STRINGP(msg)), msg, ARG4, 
	 s_xmt_display_message);
  GETARG(title, args);
  ASSERT(title == BOOL_F || (NIMP(title) && STRINGP(title)), title, ARG5, 
	 s_xmt_display_message);
  GETARG(help, args);
  ASSERT(help == BOOL_F || (NIMP(help) && STRINGP(help)), help, "ARG6", 
	 s_xmt_display_message);

#define XMTSCM_NULL_OR_STRING(s) BOOLP(s) ? NULL : CHARS(s)
  switch (INUM(type))
  {
  case 1:
    XmtDisplayErrorMsg(WIDGET(w), CHARS(name),
		       XMTSCM_NULL_OR_STRING(msg),
		       XMTSCM_NULL_OR_STRING(title),
		       XMTSCM_NULL_OR_STRING(help));
    break;
  case 2:
    XmtDisplayWarningMsg(WIDGET(w), CHARS(name),
			 XMTSCM_NULL_OR_STRING(msg),
			 XMTSCM_NULL_OR_STRING(title),
			 XMTSCM_NULL_OR_STRING(help));
    break;
  case 3:
    XmtDisplayInformationMsg(WIDGET(w), CHARS(name), 
			     XMTSCM_NULL_OR_STRING(msg),
			     XMTSCM_NULL_OR_STRING(title),
			     XMTSCM_NULL_OR_STRING(help));
    break;
  default:
    XmtDisplayInformationMsg(WIDGET(w), CHARS(name), 
			     XMTSCM_NULL_OR_STRING(msg),
			     XMTSCM_NULL_OR_STRING(title),
			     XMTSCM_NULL_OR_STRING(help));
    break;
  }
  return UNSPECIFIED;
}

SCM xmt_display_busy_cursor(w, busy)
SCM w, busy;
{
  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_display_busy_cursor);
  ASSERT(busy == BOOL_T || busy == BOOL_F, busy, ARG2, 
	 s_xmt_display_busy_cursor);
  if (busy == BOOL_T)
    XmtDisplayBusyCursor(WIDGET(w));
  else
    XmtDisplayDefaultCursor(WIDGET(w));
  return UNSPECIFIED;
}

SCM xmt_register_improved_icons(w)
SCM w;
{
  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_register_improved_icons);
  XmtRegisterImprovedIcons(WIDGET(w), NULL);
  return UNSPECIFIED;
}

SCM xmt_chooser_get_value(w, type)
SCM w, type;
{
  XtPointer val;
  extern SCM xt_make_value();

  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_chooser_get_value);
  ASSERT(NIMP(type) && SYMBOLP(type), type, ARG2, s_xmt_chooser_get_value);
  val = XmtChooserGetValue(WIDGET(w));
  return xt_make_value(type, *(XtArgVal*)val, EOL);
}

SCM xmt_block(w)
SCM w;
{
  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_block);
  blocking = True;
  XmtBlock(WIDGET(w), &blocking);
  return *block_result;
}

SCM xmt_unblock(data)
SCM data;
{
  blocking = False;
  *block_result = data;
  return UNSPECIFIED;
}

SCM xmt_create_or_change_pixmap_icon(w, pix, args)
SCM w, pix, args;
{
  Pixmap shape;
  SCM create;

  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_create_or_change_pixmap_icon);
  ASSERT(NIMP(pix) && XPIXMAPP(pix), pix, ARG2,
	 s_xmt_create_or_change_pixmap_icon);

  GETARG(shape, args);
  ASSERT(shape == BOOL_F || (NIMP(shape) && XPIXMAPP(shape)), shape, ARG3, 
	 s_xmt_create_or_change_pixmap_icon);

  ASRTARG(create, args, ARG4, s_xmt_create_or_change_pixmap_icon);
  ASSERT(BOOLP(create), create, ARG4, s_xmt_create_or_change_pixmap_icon);
  
  if (create == BOOL_T) {
/*       printf("creating pixmap icon: w=[%s], shape=%d, boolf=%d\n", */
/* 	     XtName(WIDGET(w)), shape, BOOL_F); */
    XmtCreatePixmapIcon(WIDGET(w), XPIXMAP(pix),
			BOOLCVT(shape, XPIXMAP(shape), None));
  }
  else
    XmtChangePixmapIcon(WIDGET(w), XPIXMAP(pix),
			BOOLCVT(shape, XPIXMAP(shape), None));

  return UNSPECIFIED;
}

SCM xmt_wait_until_mapped(w)
SCM w;
{
  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_wait_until_mapped);
  XmtWaitUntilMapped(WIDGET(w));
  return UNSPECIFIED;
}

static int _xmt_msg_line_when(o, argn, doc)
SCM o;
char *argn, *doc;
{
  int when = XmtMsgLineNow;		/* default -> clear now */

  ASSERT(INUMP(o) || BOOLP(o),  o, argn, doc);
  if (o == BOOL_F) { 
    when = XmtMsgLineOnAction;	/* false -> on action */
  } else if (INUMP(o)) { 
    when = INUM(o);			/* integer -> time interval */
  }
  return when;
}

SCM xmt_msg_line_set(w, s) 
SCM w, s;
{
  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_msg_line_set);
  ASSERT(NIMP(s) && STRINGP(s), s, ARG2, s_xmt_msg_line_set);
  XmtMsgLineSet(WIDGET(w), CHARS(s));
  return UNSPECIFIED;
}

SCM xmt_msg_line_append(w, s) 
SCM w, s;
{
  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_msg_line_append);
  ASSERT(NIMP(s) && STRINGP(s), s, ARG2, s_xmt_msg_line_append);
  XmtMsgLineAppend(WIDGET(w), CHARS(s));
  return UNSPECIFIED;
}

SCM xmt_msg_line_clear(w, o) 
SCM w, o;
{
  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_msg_line_clear);
  XmtMsgLineClear(WIDGET(w), 
		  _xmt_msg_line_when(o, ARG2, s_xmt_msg_line_clear));
  return UNSPECIFIED;
}

SCM xmt_msg_line_push(w)
SCM w;
{
  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_msg_line_push);
  XmtMsgLinePush(WIDGET(w));
  return UNSPECIFIED;
}

SCM xmt_msg_line_pop(w, o)
SCM w, o;
{
  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_msg_line_pop);
  XmtMsgLinePop(WIDGET(w),
		_xmt_msg_line_when(o, ARG2, s_xmt_msg_line_pop));
  return UNSPECIFIED;
}

SCM xmt_msg_line_get(w, type)
SCM w, type;
{
  SCM result = BOOL_F;
  char buf[256], *s;
  int c;
  unsigned u;
  double d;

  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_msg_line_get);
  ASSERT(INUMP(type), type,        ARG2, s_xmt_msg_line_get);
  switch (INUM(type))
  {
  case 0:				/* string */
    s = XmtMsgLineGetString(WIDGET(w), buf, sizeof(buf));
    if (s) { result = makfrom0str(s); }
    break;
  case 1:				/* character */
    c = XmtMsgLineGetChar(WIDGET(w));
    if (c != EOF) { result = MAKICHR(c); }
    break;
  case 2:				/* unsigned */
    if (XmtMsgLineGetUnsigned(WIDGET(w), &u))
      result = MAKINUM(u);
    break;
  case 3:				/* integer */
    if (XmtMsgLineGetInt(WIDGET(w), &c))
      result = MAKINUM(c);
    break;
  case 4:				/* double */
    if (XmtMsgLineGetDouble(WIDGET(w), &d))
      result = makdbl(d, 0.0);
    break;
  }
  return result;
}

SCM xmt_menu_get_widget(w, s)
SCM w, s;
{
  XmtMenuItem *item;
  Widget x = 0;

  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_menu_get_widget);
  ASSERT(NIMP(s) && STRINGP(s), s, ARG2, s_xmt_menu_get_widget);
  item = XmtMenuGetMenuItem(WIDGET(w), CHARS(s));
  if (item) x = XmtMenuItemGetWidget(item);

  return x ? make_widget(x) : BOOL_F;
}

extern void xt__make_arglist();

SCM xmt_set_application_values(args)
SCM args;
{
  SCM w;
  ArgList arglist;
  int n;

  ASSERT(NIMP(args) && CONSP(args), args, ARG1, s_xmt_set_application_values);
  w = CAR(args); args = CDR(args);
  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_set_application_values);
  xt__make_arglist(args, &arglist, &n, s_xmt_set_application_values);
  XmtSetApplicationValues(WIDGET(w), arglist, n);
  return UNSPECIFIED;
}

SCM xmt_get_bitpixmap(sw, sn, sb)
SCM sw, sn, sb;
{
  SCM sp;
  Pixmap p;

  ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xmt_get_bitpixmap);
  ASSERT(NIMP(sn) && STRINGP(sn), sn, ARG2, s_xmt_get_bitpixmap);
  ASSERT(BOOLP(sb), sb, ARG3, s_xmt_get_bitpixmap);
  p = (sb == BOOL_T) 
    ? XmtGetPixmap(WIDGET(sw), NULL, CHARS(sn))
    : XmtGetBitmap(WIDGET(sw), CHARS(sn));
  if (p == None) 
    return BOOL_F;
  sp = make_xpixmap();
  SETCDR(sp, p);
  return sp;
}

SCM xmt_release_pixmap(sw, sp)
SCM sw, sp;
{
  ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xmt_release_pixmap);
  ASSERT(NIMP(sp) && XPIXMAPP(sp), sp, ARG2, s_xmt_release_pixmap);
  XmtReleasePixmap(WIDGET(sw), XPIXMAP(sp));
  return UNSPECIFIED;
}

SCM xmt_manipulate_shell(sw, val)
SCM sw, val;
{
  ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xmt_manipulate_shell);
  ASSERT(INUMP(val) && INUM(val)>=1 && INUM(val)<=4, val, ARG2, 
	 s_xmt_manipulate_shell);
  switch (INUM(val))
  {
  case 1: XmtIconifyShell(WIDGET(sw));   break;
  case 2: XmtDeiconifyShell(WIDGET(sw)); break;
  case 3: XmtRaiseShell(WIDGET(sw));     break;
  case 4: XmtLowerShell(WIDGET(sw));     break;
  }
  return UNSPECIFIED;
}

SCM xmt_dialog_position(d, ref)
SCM d, ref;
{
  ASSERT(NIMP(d) && WIDGETP(d), d, ARG1, s_xmt_dialog_position);
  ASSERT(NIMP(ref) && WIDGETP(ref), ref, ARG2, s_xmt_dialog_position);
  XmtDialogPosition(WIDGET(d), WIDGET(ref));
  return UNSPECIFIED;
}

void xmt_wm_callback_handler(w, p, q)
Widget w;
XtPointer p, q;
{
  SCM args, data = (SCM)p;

  args = cons(CDR(data), EOL);
  args = cons(args, EOL);
  apply(CAR(data), make_widget(w), args);
}

SCM xmt_add_wm_callback(sw, type, args)
SCM sw, type, args;
{
  SCM response, proc, data = BOOL_F, pass;

  ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xmt_add_wm_callback);
  ASSERT(BOOLP(type), type, ARG2, s_xmt_add_wm_callback);

  ASRTARG(response, args, ARG3, s_xmt_add_wm_callback);
  ASSERT(INUMP(response), response, ARG3, s_xmt_add_wm_callback);

  ASRTARG(proc, args, ARG4, s_xmt_add_wm_callback);
  ASSERT(NIMP(proc) && CLOSUREP(proc), proc, ARG4, s_xmt_add_wm_callback);

  ASRTARG(data, args, ARG5, s_xmt_add_wm_callback);
  pass = cons(proc, data);
  protect_callback(pass);

  if (type == BOOL_T)			/* delete callback */
  {
    XmtAddDeleteCallback(XmtGetShell(WIDGET(sw)),
			 INUM(response),
			 xmt_wm_callback_handler,
			 (XtPointer)pass);
  } 
  else					/* save yourself callback */
  {
    XmtAddSaveYourselfCallback(XmtGetShell(WIDGET(sw)),
			       xmt_wm_callback_handler,
			       (XtPointer)pass);
  }
  return UNSPECIFIED;
}

SCM xmt_layout_xable(sw, opt)
SCM sw, opt;
{
  ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xmt_layout_xable);
  ASSERT(BOOLP(opt), opt, ARG2, s_xmt_layout_xable);
  if (opt == BOOL_T)
    XmtLayoutEnableLayout(WIDGET(sw));
  else
    XmtLayoutDisableLayout(WIDGET(sw));
  return UNSPECIFIED;
}

SCM xmt_parse_xpm_file(file)
SCM file;
{
  XmtImage *img;

  ASSERT(NIMP(file) && STRINGP(file), file, ARG1, s_xmt_parse_xpm_file);
  if ((img = XmtParseXpmFile(CHARS(file)))) 
    return make_xmtimage(img);
  else
    return BOOL_F;
}

SCM xmt_get_xmt_image_field(img, fld)
SCM img, fld;
{
  ASSERT(NIMP(img) && XMTIMAGE(img), img, ARG1, s_xmt_get_xmt_image_field);
  ASSERT(INUMP(fld),                 fld, ARG2, s_xmt_get_xmt_image_field);
  switch (fld) 
  {
  case 0: return MAKINUM(XMTIMAGE(img)->width);
  case 1: return MAKINUM(XMTIMAGE(img)->height);
  case 2: return cons(MAKINUM(XMTIMAGE(img)->hotspot.x), 
		      MAKINUM(XMTIMAGE(img)->hotspot.y));
  case 3: return VALBOOL(XMTIMAGE(img)->has_hotspot);
  case 4: return MAKINUM(XMTIMAGE(img)->ncolors);
  case 5: return makfromstr((char*)XMTIMAGE(img)->data, 
			    XMTIMAGE(img)->width * XMTIMAGE(img)->height);
  default:
    wta(fld, "Wrong field type", s_xmt_get_xmt_image_field);
  }
  /*unreachable*/
  return UNSPECIFIED;
}

SCM xmt_create_pixmap_from_xmt_image(sw, win, img)
SCM sw, win, img;
{
  Pixmap pix, mask;

  ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xmt_create_pixmap_from_xmt_image);
  ASSERT(NIMP(win) && XWINDOWP(win), win, ARG2, s_xmt_create_pixmap_from_xmt_image);
  ASSERT(NIMP(img) && XMTIMAGEP(img), img, ARG3, s_xmt_create_pixmap_from_xmt_image);

  if (XmtCreatePixmapFromXmtImage(WIDGET(sw), XWINDOW(win), NULL, None,
				  0, NULL, XMTIMAGE(img), 
				  &pix, &mask, NULL, NULL))
    return cons(pix, mask);
  else
    return BOOL_F;
}

SCM xmt_load_resource_file(w, f, args) /* w=widget, f=filename */
SCM w, f, args;
{
  SCM u, o;			/* u=user, o=override */

  ASSERT(NIMP(w) && WIDGETP(w), w, ARG1, s_xmt_load_resource_file);
  ASSERT(NIMP(f) && STRINGP(f), f, ARG2, s_xmt_load_resource_file);
  ASRTARG(u, args, ARG3, s_xmt_load_resource_file);
  ASSERT(BOOLP(u), u, ARG3, s_xmt_load_resource_file);
  ASRTARG(o, args, ARG3, s_xmt_load_resource_file);
  ASSERT(BOOLP(o), o, ARG4, s_xmt_load_resource_file);

  return VALBOOL(XmtLoadResourceFile(WIDGET(w), CHARS(f),
				     BOOLVAL(u), BOOLVAL(o)));
}



#ifdef USE_RESOURCE_MANAGER
SCM xrm_put_line_resource(sw, line)
SCM sw, line;
{
  XrmDatabase db;

  ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xrm_put_line_resource);
  ASSERT(NIMP(line) && STRINGP(line), line, ARG2, s_xrm_put_line_resource);
  db = XrmGetDatabase(XtDisplay(WIDGET(sw)));
  XrmPutLineResource(&db, CHARS(line));
  XrmSetDatabase(XtDisplay(WIDGET(sw)), db);
  return UNSPECIFIED;
}
#endif

iproc xmt_subr1s[] = {
  {s_xmt_create_children,	xmt_create_children},
  {s_xmt_register_improved_icons, xmt_register_improved_icons},
  {s_xmt_block,                 xmt_block},
  {s_xmt_unblock,               xmt_unblock},
  {s_xmt_wait_until_mapped,     xmt_wait_until_mapped},
  {s_xmt_msg_line_push,         xmt_msg_line_push},
  {s_xmt_parse_xpm_file,        xmt_parse_xpm_file},
  {0, 0}
};

iproc xmt_subr2s[] = {
  {s_xmt_initialize,		xmt_initialize},
  {s_xmt_name_to_widget,	xmt_name_to_widget},
  {s_xmt_display_busy_cursor,   xmt_display_busy_cursor},
  {s_xmt_generic_get_shell,     xmt_generic_get_shell},
  {s_xmt_chooser_get_value,     xmt_chooser_get_value},
  {s_xmt_msg_line_set,          xmt_msg_line_set},
  {s_xmt_msg_line_append,       xmt_msg_line_append},
  {s_xmt_msg_line_clear,        xmt_msg_line_clear},
  {s_xmt_msg_line_pop,          xmt_msg_line_pop},
  {s_xmt_msg_line_get,          xmt_msg_line_get},
  {s_xmt_menu_get_widget,       xmt_menu_get_widget},
  {s_xmt_manipulate_shell,      xmt_manipulate_shell},
  {s_xmt_dialog_position,       xmt_dialog_position},
  {s_xmt_release_pixmap,        xmt_release_pixmap},
  {s_xmt_layout_xable,          xmt_layout_xable},
  {s_xmt_get_xmt_image_field,   xmt_get_xmt_image_field},
#ifdef USE_RESOURCE_MANAGER
  {s_xrm_put_line_resource,     xrm_put_line_resource},
#endif
  {0, 0}
};

iproc xmt_subr3s[] = {
  {s_xmt_ask_for_string,	xmt_ask_for_string},
  {s_xmt_get_bitpixmap,         xmt_get_bitpixmap},
  {s_xmt_create_pixmap_from_xmt_image, xmt_create_pixmap_from_xmt_image},
  {0, 0}
};

iproc xmt_lsubrs[] = {
  {s_xmt_set_application_values, xmt_set_application_values},
  {0,0}
};
  
iproc xmt_lsubr2s[] = {
  {s_xmt_ask_for_boolean,	xmt_ask_for_boolean},
  {s_xmt_display_message,	xmt_display_message},
  {s_xmt_ask_for_filename,	xmt_ask_for_filename},
  {s_xmt_ask_for_item,		xmt_ask_for_item},
  {s_xmt_add_wm_callback,       xmt_add_wm_callback},
  {s_xmt_create_or_change_pixmap_icon, xmt_create_or_change_pixmap_icon},
  {s_xmt_load_resource_file,    xmt_load_resource_file},
  {0, 0}
};

void init_xmt()
{
  SCM s;

  x_make_integer_constants(xmt_symbols);
  INTERN(s, "*xmt-block-result*", EOL);
  block_result = &CDR(s);

  tc16_xmtimage = newsmob(&xmtimage_smob);
  init_iprocs(xmt_subr1s,  tc7_subr_1);
  init_iprocs(xmt_subr2s,  tc7_subr_2);
  init_iprocs(xmt_subr3s,  tc7_subr_3);
  init_iprocs(xmt_lsubrs,  tc7_lsubr);
  init_iprocs(xmt_lsubr2s, tc7_lsubr_2);
  xt_init_widget_classes(
    xmt_widget_classes,
    XtNumber(xmt_widget_classes),
    "*xmt-motif-widget-classes*");
}
